import matplotlib.pyplot as plt
from matplotlib.lines import Line2D
import matplotlib.ticker as mticker
import numpy as np
import pandas as pd
import plotly.graph_objects as go
from .constants import plot_2_x, plot_2_y


class CarbonBudgetAssessmentPlot:
    def __init__(self, data):
        self.parameters = data["float_inputs"]
        self.df = data["vector_outputs"]
        self.float_outputs = data["float_outputs"]
        self.years = data["years"]["full_years"]
        self.historic_years = data["years"]["historic_years"]
        self.prospective_years = data["years"]["prospective_years"]

        self.fig, self.ax = plt.subplots(
            figsize=(plot_2_x, plot_2_y),
        )
        self.create_plot()

    def create_plot(self):
        # Data to plot
        aviation_carbon_budget = float(self.float_outputs["aviation_carbon_budget"])
        global_cumulative_co2_emissions_2050 = float(self.df.loc[2050, "cumulative_co2_emissions"])
        world_carbon_budget = float(self.float_outputs["gross_carbon_budget_2050"])
        aviation_carbon_budget_allocated_share = float(
            self.parameters["aviation_carbon_budget_allocated_share"]
        )

        # Plot
        size = 0.3
        outer_vals = [
            global_cumulative_co2_emissions_2050,
            world_carbon_budget - global_cumulative_co2_emissions_2050,
        ]
        inner_vals = [aviation_carbon_budget, world_carbon_budget - aviation_carbon_budget]

        color = "skyblue"

        if global_cumulative_co2_emissions_2050 <= aviation_carbon_budget:
            color = "green"
        outer_colors = [color, "white"]
        inner_colors = ["grey", "white"]

        self.wedges_out, _ = self.ax.pie(
            outer_vals,
            radius=1,
            colors=outer_colors,
            startangle=90,
            wedgeprops=dict(width=size, edgecolor="k", alpha=0.8),
        )

        self.wedges_in, _ = self.ax.pie(
            inner_vals,
            radius=1 - size,
            colors=inner_colors,
            startangle=90,
            wedgeprops=dict(width=size, edgecolor="k", alpha=0.8),
        )

        bbox_props = dict(boxstyle="square,pad=0.3", fc="w", ec="k", lw=0.72)
        kw = dict(arrowprops=dict(arrowstyle="-"), bbox=bbox_props, zorder=0, va="center")

        self.ax.set(
            aspect="equal",
            title="Estimated climate impact of aviation until 2050\nPresented as a % of the world carbon budget\n(WCB) 2050",
        )

        wedges = [self.wedges_out[0], self.wedges_in[0]]
        texts = [
            str(round(global_cumulative_co2_emissions_2050, 1))
            + " GtCO2\ncorresponding to\n "
            + str(round(global_cumulative_co2_emissions_2050 / world_carbon_budget * 100, 1))
            + "% of WCB",
            str(round(aviation_carbon_budget_allocated_share, 1))
            + "% of WCB\ni.e.\n"
            + str(round(aviation_carbon_budget, 1))
            + " GtCO2",
        ]

        p = wedges[0]
        text = texts[0]

        ang = (p.theta2 - p.theta1) / 2.0 + p.theta1
        y = np.sin(np.deg2rad(ang))
        x = np.cos(np.deg2rad(ang))
        horizontalalignment = {-1: "right", 1: "left"}[int(np.sign(x))]
        connectionstyle = "angle,angleA=0,angleB={}".format(ang)
        kw["arrowprops"].update({"connectionstyle": connectionstyle})
        self.ax.annotate(
            text,
            xy=(x, y),
            xytext=(1.2 * np.sign(x), 1.1 * y),
            horizontalalignment=horizontalalignment,
            **kw,
        )

        p = wedges[1]
        text = texts[1]

        ang = (p.theta2 - p.theta1) / 2.0 + p.theta1 - 45
        y = np.sin(np.deg2rad(ang)) * 0.7
        x = np.cos(np.deg2rad(ang)) * 0.00001
        horizontalalignment = {-1: "right", 1: "left"}[int(np.sign(x))]
        connectionstyle = "angle,angleA=0,angleB={}".format(ang)
        kw["arrowprops"].update({"connectionstyle": connectionstyle, "facecolor": "black"})
        self.ax.annotate(
            text,
            xy=(-x, y),
            xytext=(1.2 * np.sign(x), 2.0 * y),
            horizontalalignment=horizontalalignment,
            **kw,
        )

        legend_elements = [
            Line2D(
                [0],
                [0],
                color=color,
                lw=6,
                label="Cumulative CO2 emissions of aviation\nbetween 2020 and 2050",
            ),
            Line2D(
                [0],
                [0],
                color="grey",
                lw=6,
                label="Carbon budget 2050 allocated to aviation",
            ),
        ]

        self.ax.legend(handles=legend_elements, loc="center", bbox_to_anchor=[0.5, -0.12])

        self.fig.canvas.header_visible = False
        self.fig.canvas.toolbar_position = "bottom"
        # self.fig.canvas.layout.width = "auto"
        # self.fig.canvas.layout.height = "auto"
        # self.fig.set_tight_layout(True)

        self.fig.tight_layout()

    def update(self, data):
        self.parameters = data["float_inputs"]
        self.df = data["vector_outputs"]
        self.float_outputs = data["float_outputs"]
        self.years = data["years"]["full_years"]
        self.historic_years = data["years"]["historic_years"]
        self.prospective_years = data["years"]["prospective_years"]

        self.ax.clear()

        aviation_carbon_budget = float(self.float_outputs["aviation_carbon_budget"])
        global_cumulative_co2_emissions_2050 = float(self.df.loc[2050, "cumulative_co2_emissions"])
        world_carbon_budget = float(self.float_outputs["gross_carbon_budget_2050"])
        aviation_carbon_budget_allocated_share = float(
            self.parameters["aviation_carbon_budget_allocated_share"]
        )

        # Plot
        size = 0.3
        outer_vals = [
            global_cumulative_co2_emissions_2050,
            world_carbon_budget - global_cumulative_co2_emissions_2050,
        ]
        inner_vals = [aviation_carbon_budget, world_carbon_budget - aviation_carbon_budget]

        color = "skyblue"

        if global_cumulative_co2_emissions_2050 <= aviation_carbon_budget:
            color = "green"
        outer_colors = [color, "white"]
        inner_colors = ["grey", "white"]

        self.wedges_out, _ = self.ax.pie(
            outer_vals,
            radius=1,
            colors=outer_colors,
            startangle=90,
            wedgeprops=dict(width=size, edgecolor="k", alpha=0.8),
        )

        self.wedges_in, _ = self.ax.pie(
            inner_vals,
            radius=1 - size,
            colors=inner_colors,
            startangle=90,
            wedgeprops=dict(width=size, edgecolor="k", alpha=0.8),
        )

        bbox_props = dict(boxstyle="square,pad=0.3", fc="w", ec="k", lw=0.72)
        kw = dict(arrowprops=dict(arrowstyle="-"), bbox=bbox_props, zorder=0, va="center")

        self.ax.set(
            aspect="equal",
            title="Estimated climate impact of aviation until 2050\nPresented as a % of the world carbon budget\n(WCB) 2050",
        )

        wedges = [self.wedges_out[0], self.wedges_in[0]]
        texts = [
            str(round(global_cumulative_co2_emissions_2050, 1))
            + " GtCO2\ncorresponding to\n "
            + str(round(global_cumulative_co2_emissions_2050 / world_carbon_budget * 100, 1))
            + "% of WCB",
            str(round(aviation_carbon_budget_allocated_share, 1))
            + "% of WCB\ni.e.\n"
            + str(round(aviation_carbon_budget, 1))
            + " GtCO2",
        ]

        p = wedges[0]
        text = texts[0]

        ang = (p.theta2 - p.theta1) / 2.0 + p.theta1
        y = np.sin(np.deg2rad(ang))
        x = np.cos(np.deg2rad(ang))
        horizontalalignment = {-1: "right", 1: "left"}[int(np.sign(x))]
        connectionstyle = "angle,angleA=0,angleB={}".format(ang)
        kw["arrowprops"].update({"connectionstyle": connectionstyle})
        self.ax.annotate(
            text,
            xy=(x, y),
            xytext=(1.2 * np.sign(x), 1.1 * y),
            horizontalalignment=horizontalalignment,
            **kw,
        )

        p = wedges[1]
        text = texts[1]

        ang = (p.theta2 - p.theta1) / 2.0 + p.theta1 - 45
        y = np.sin(np.deg2rad(ang)) * 0.7
        x = np.cos(np.deg2rad(ang)) * 0.00001
        horizontalalignment = {-1: "right", 1: "left"}[int(np.sign(x))]
        connectionstyle = "angle,angleA=0,angleB={}".format(ang)
        kw["arrowprops"].update({"connectionstyle": connectionstyle, "facecolor": "black"})
        self.ax.annotate(
            text,
            xy=(-x, y),
            xytext=(1.2 * np.sign(x), 2.0 * y),
            horizontalalignment=horizontalalignment,
            **kw,
        )

        legend_elements = [
            Line2D(
                [0],
                [0],
                color=color,
                lw=6,
                label="Cumulative CO2 emissions of aviation\nbetween 2020 and 2050",
            ),
            Line2D(
                [0],
                [0],
                color="grey",
                lw=6,
                label="Carbon budget 2050 allocated to aviation",
            ),
        ]

        self.ax.legend(handles=legend_elements, loc="center", bbox_to_anchor=[0.5, -0.12])

        self.fig.canvas.draw()


class EquivalentCarbonBudgetAssessmentPlot:
    def __init__(self, data):
        self.parameters = data["float_inputs"]
        self.df = data["vector_outputs"]
        self.float_outputs = data["float_outputs"]
        self.years = data["years"]["full_years"]
        self.historic_years = data["years"]["historic_years"]
        self.prospective_years = data["years"]["prospective_years"]

        self.fig, self.ax = plt.subplots(
            figsize=(plot_2_x, plot_2_y),
        )
        self.create_plot()

    def create_plot(self):
        barWidth = 0.6

        y1 = [
            self.df.loc[2050, "cumulative_total_equivalent_emissions"],
            self.float_outputs["aviation_equivalent_carbon_budget"],
        ]
        r = range(len(y1))
        self.line_carbon_budget = self.ax.bar(r, y1, width=barWidth, color=["skyblue" for i in y1])

        i = 0
        for rect, h in zip(self.line_carbon_budget, y1):
            rect.set_height(h)
            if i == 0:
                if y1[i] < y1[i + 1]:
                    rect.set_color("green")
                else:
                    rect.set_color("skyblue")
            i += 1

        self.ax.set_xticks(r)
        self.ax.set_xticklabels(
            [
                "Equivalent cumulative\nemissions",
                "Aviation equivalent\naviation carbon\nbudget 2050",
            ],
        )
        self.ax.set_title(
            "Estimated climate impact of aviation until 2050\nPresented in relation to equivalent\ncarbon budget for aviation"
        )
        self.ax.set_ylabel("Equivalent cumulative emissions\nbetween 2020 and 2050 [GtCO2-we]")

        self.ax2 = self.ax.twinx()

        ymin, ymax = self.ax.get_ylim()
        self.ax2.set_ylim([ymin, ymax])
        new_labels = [
            round(float(self.float_outputs["equivalent_carbon_budget_consumed_share"]), 1),
            round(self.parameters["aviation_equivalentcarbonbudget_allocated_share"], 1),
        ]

        self.ax2.set_yticks([y1[0], float(y1[1])])
        self.ax2.set_yticklabels(new_labels)
        self.ax2.set_ylabel("Share of the world equivalent\ncarbon budget [%]")

        self.fig.canvas.header_visible = False
        self.fig.canvas.toolbar_position = "bottom"
        # self.fig.canvas.layout.width = "auto"
        # self.fig.canvas.layout.height = "auto"
        self.fig.tight_layout()

    def update(self, data):
        self.parameters = data["float_inputs"]
        self.df = data["vector_outputs"]
        self.float_outputs = data["float_outputs"]
        self.years = data["years"]["full_years"]
        self.historic_years = data["years"]["historic_years"]
        self.prospective_years = data["years"]["prospective_years"]

        y1 = [
            self.df.loc[2050, "cumulative_total_equivalent_emissions"],
            self.float_outputs["aviation_equivalent_carbon_budget"],
        ]
        i = 0
        for rect, h in zip(self.line_carbon_budget, y1):
            rect.set_height(h)
            if i == 0:
                if y1[i] < y1[i + 1]:
                    rect.set_color("green")
                else:
                    rect.set_color("skyblue")
            i += 1

        self.ax.relim()
        self.ax.autoscale_view()

        ymin, ymax = self.ax.get_ylim()
        self.ax2.set_ylim([ymin, ymax])
        new_labels = [
            round(float(self.float_outputs["equivalent_carbon_budget_consumed_share"]), 1),
            round(self.parameters["aviation_equivalentcarbonbudget_allocated_share"], 1),
        ]
        self.ax2.set_yticks([y1[0], float(y1[1])])
        self.ax2.set_yticklabels(new_labels)

        self.fig.canvas.draw()


class BiomassResourceBudgetAssessmentPlot:
    def __init__(self, data):
        self.parameters = data["float_inputs"]
        self.df = data["vector_outputs"]
        self.float_outputs = data["float_outputs"]
        self.years = data["years"]["full_years"]
        self.historic_years = data["years"]["historic_years"]
        self.prospective_years = data["years"]["prospective_years"]

        self.fig, self.ax = plt.subplots(
            figsize=(plot_2_x, plot_2_y),
        )
        self.create_plot()

    def create_plot(self):
        # Data to plot
        aviation_available_biomass = float(self.float_outputs["aviation_available_biomass"])
        biomass_consumption_end_year = float(self.float_outputs["biomass_consumption_end_year"])
        available_biomass_total = float(self.float_outputs["available_biomass_total"])
        aviation_biomass_allocated_share = float(
            self.parameters["aviation_biomass_allocated_share"]
        )

        # Plot
        size = 0.3
        outer_vals = [
            biomass_consumption_end_year,
            available_biomass_total - biomass_consumption_end_year,
        ]
        inner_vals = [
            aviation_available_biomass,
            available_biomass_total - aviation_available_biomass,
        ]

        color = "skyblue"

        if biomass_consumption_end_year <= aviation_available_biomass:
            color = "green"
        outer_colors = [color, "white"]
        inner_colors = ["grey", "white"]

        self.wedges_out, _ = self.ax.pie(
            outer_vals,
            radius=1,
            colors=outer_colors,
            startangle=90,
            wedgeprops=dict(width=size, edgecolor="k", alpha=0.8),
        )

        self.wedges_in, _ = self.ax.pie(
            inner_vals,
            radius=1 - size,
            colors=inner_colors,
            startangle=90,
            wedgeprops=dict(width=size, edgecolor="k", alpha=0.8),
        )

        bbox_props = dict(boxstyle="square,pad=0.3", fc="w", ec="k", lw=0.72)
        kw = dict(arrowprops=dict(arrowstyle="-"), bbox=bbox_props, zorder=0, va="center")

        self.ax.set(
            aspect="equal",
            title="Estimated biomass consumption of aviation in 2050\nPresented as a % of the available biomass\n(AB) in 2050",
        )

        wedges = [self.wedges_out[0], self.wedges_in[0]]
        texts = [
            str(round(biomass_consumption_end_year, 1))
            + " EJ\ncorresponding to\n "
            + str(round(biomass_consumption_end_year / available_biomass_total * 100, 1))
            + "% of AB",
            str(round(aviation_biomass_allocated_share, 1))
            + "% of AB\ni.e.\n"
            + str(round(aviation_available_biomass, 1))
            + " EJ",
        ]

        p = wedges[0]
        text = texts[0]

        ang = (p.theta2 - p.theta1) / 2.0 + p.theta1
        y = np.sin(np.deg2rad(ang))
        x = np.cos(np.deg2rad(ang))
        horizontalalignment = {-1: "right", 1: "left"}[int(np.sign(x))]
        connectionstyle = "angle,angleA=0,angleB={}".format(ang)
        kw["arrowprops"].update({"connectionstyle": connectionstyle})
        self.ax.annotate(
            text,
            xy=(x, y),
            xytext=(1.2 * np.sign(x), 1.1 * y),
            horizontalalignment=horizontalalignment,
            **kw,
        )

        p = wedges[1]
        text = texts[1]

        ang = (p.theta2 - p.theta1) / 2.0 + p.theta1 - 45
        y = np.sin(np.deg2rad(ang)) * 0.7
        x = np.cos(np.deg2rad(ang)) * 0.00001
        horizontalalignment = {-1: "right", 1: "left"}[int(np.sign(x))]
        connectionstyle = "angle,angleA=0,angleB={}".format(ang)
        kw["arrowprops"].update({"connectionstyle": connectionstyle, "facecolor": "black"})
        self.ax.annotate(
            text,
            xy=(-x, y),
            xytext=(1.2 * np.sign(x), 2.0 * y),
            horizontalalignment=horizontalalignment,
            **kw,
        )

        legend_elements = [
            Line2D(
                [0],
                [0],
                color=color,
                lw=6,
                label="Biomass consumption of aviation in 2050",
            ),
            Line2D(
                [0],
                [0],
                color="grey",
                lw=6,
                label="Biomass resources allocated to aviation",
            ),
        ]

        self.ax.legend(handles=legend_elements, loc="center", bbox_to_anchor=[0.5, -0.12])

        self.fig.canvas.header_visible = False
        self.fig.canvas.toolbar_position = "bottom"
        # self.fig.canvas.layout.width = "auto"
        # self.fig.canvas.layout.height = "auto"
        # self.fig.set_tight_layout(True)

        self.fig.tight_layout()

    def update(self, data):
        self.parameters = data["float_inputs"]
        self.df = data["vector_outputs"]
        self.float_outputs = data["float_outputs"]
        self.years = data["years"]["full_years"]
        self.historic_years = data["years"]["historic_years"]
        self.prospective_years = data["years"]["prospective_years"]

        self.ax.clear()

        aviation_available_biomass = float(self.float_outputs["aviation_available_biomass"])
        biomass_consumption_end_year = float(self.float_outputs["biomass_consumption_end_year"])
        available_biomass_total = float(self.float_outputs["available_biomass_total"])
        aviation_biomass_allocated_share = float(
            self.parameters["aviation_biomass_allocated_share"]
        )

        # Plot
        size = 0.3
        outer_vals = [
            biomass_consumption_end_year,
            available_biomass_total - biomass_consumption_end_year,
        ]
        inner_vals = [
            aviation_available_biomass,
            available_biomass_total - aviation_available_biomass,
        ]

        color = "skyblue"

        if biomass_consumption_end_year <= aviation_available_biomass:
            color = "green"
        outer_colors = [color, "white"]
        inner_colors = ["grey", "white"]

        self.wedges_out, _ = self.ax.pie(
            outer_vals,
            radius=1,
            colors=outer_colors,
            startangle=90,
            wedgeprops=dict(width=size, edgecolor="k", alpha=0.8),
        )

        self.wedges_in, _ = self.ax.pie(
            inner_vals,
            radius=1 - size,
            colors=inner_colors,
            startangle=90,
            wedgeprops=dict(width=size, edgecolor="k", alpha=0.8),
        )

        bbox_props = dict(boxstyle="square,pad=0.3", fc="w", ec="k", lw=0.72)
        kw = dict(arrowprops=dict(arrowstyle="-"), bbox=bbox_props, zorder=0, va="center")

        self.ax.set(
            aspect="equal",
            title="Estimated biomass consumption of aviation in 2050\nPresented as a % of the available biomass\n(AB) in 2050",
        )

        wedges = [self.wedges_out[0], self.wedges_in[0]]
        texts = [
            str(round(biomass_consumption_end_year, 1))
            + " EJ\ncorresponding to\n "
            + str(round(biomass_consumption_end_year / available_biomass_total * 100, 1))
            + "% of AB",
            str(round(aviation_biomass_allocated_share, 1))
            + "% of AB\ni.e.\n"
            + str(round(aviation_available_biomass, 1))
            + " EJ",
        ]

        p = wedges[0]
        text = texts[0]

        ang = (p.theta2 - p.theta1) / 2.0 + p.theta1
        y = np.sin(np.deg2rad(ang))
        x = np.cos(np.deg2rad(ang))
        horizontalalignment = {-1: "right", 1: "left"}[int(np.sign(x))]
        connectionstyle = "angle,angleA=0,angleB={}".format(ang)
        kw["arrowprops"].update({"connectionstyle": connectionstyle})
        self.ax.annotate(
            text,
            xy=(x, y),
            xytext=(1.2 * np.sign(x), 1.1 * y),
            horizontalalignment=horizontalalignment,
            **kw,
        )

        p = wedges[1]
        text = texts[1]

        ang = (p.theta2 - p.theta1) / 2.0 + p.theta1 - 45
        y = np.sin(np.deg2rad(ang)) * 0.7
        x = np.cos(np.deg2rad(ang)) * 0.00001
        horizontalalignment = {-1: "right", 1: "left"}[int(np.sign(x))]
        connectionstyle = "angle,angleA=0,angleB={}".format(ang)
        kw["arrowprops"].update({"connectionstyle": connectionstyle, "facecolor": "black"})
        self.ax.annotate(
            text,
            xy=(-x, y),
            xytext=(1.2 * np.sign(x), 2.0 * y),
            horizontalalignment=horizontalalignment,
            **kw,
        )

        legend_elements = [
            Line2D(
                [0],
                [0],
                color=color,
                lw=6,
                label="Biomass consumption of aviation in 2050",
            ),
            Line2D(
                [0],
                [0],
                color="grey",
                lw=6,
                label="Biomass resources allocated to aviation",
            ),
        ]

        self.ax.legend(handles=legend_elements, loc="center", bbox_to_anchor=[0.5, -0.12])

        self.fig.canvas.draw()


class ElectricityResourceBudgetAssessmentPlot:
    def __init__(self, data):
        self.parameters = data["float_inputs"]
        self.df = data["vector_outputs"]
        self.float_outputs = data["float_outputs"]
        self.years = data["years"]["full_years"]
        self.historic_years = data["years"]["historic_years"]
        self.prospective_years = data["years"]["prospective_years"]

        self.fig, self.ax = plt.subplots(
            figsize=(plot_2_x, plot_2_y),
        )
        self.create_plot()

    def create_plot(self):
        # Data to plot
        aviation_available_electricity = float(self.float_outputs["aviation_available_electricity"])
        electricity_consumption_end_year = float(
            self.float_outputs["electricity_consumption_end_year"]
        )
        available_electricity = float(self.parameters["available_electricity"])
        aviation_electricity_allocated_share = float(
            self.parameters["aviation_electricity_allocated_share"]
        )

        # Plot
        size = 0.3
        outer_vals = [
            electricity_consumption_end_year,
            available_electricity - electricity_consumption_end_year,
        ]
        inner_vals = [
            aviation_available_electricity,
            available_electricity - aviation_available_electricity,
        ]

        color = "skyblue"

        if electricity_consumption_end_year <= aviation_available_electricity:
            color = "green"
        outer_colors = [color, "white"]
        inner_colors = ["grey", "white"]

        self.wedges_out, _ = self.ax.pie(
            outer_vals,
            radius=1,
            colors=outer_colors,
            startangle=90,
            wedgeprops=dict(width=size, edgecolor="k", alpha=0.8),
        )

        self.wedges_in, _ = self.ax.pie(
            inner_vals,
            radius=1 - size,
            colors=inner_colors,
            startangle=90,
            wedgeprops=dict(width=size, edgecolor="k", alpha=0.8),
        )

        bbox_props = dict(boxstyle="square,pad=0.3", fc="w", ec="k", lw=0.72)
        kw = dict(arrowprops=dict(arrowstyle="-"), bbox=bbox_props, zorder=0, va="center")

        self.ax.set(
            aspect="equal",
            title="Estimated electricity consumption of aviation in 2050\nPresented as a % of the available elecltricity\n(AE) in 2050",
        )

        wedges = [self.wedges_out[0], self.wedges_in[0]]
        texts = [
            str(round(electricity_consumption_end_year, 1))
            + " EJ\ncorresponding to\n "
            + str(round(electricity_consumption_end_year / available_electricity * 100, 1))
            + "% of AE",
            str(round(aviation_electricity_allocated_share, 1))
            + "% of AE\ni.e.\n"
            + str(round(aviation_available_electricity, 1))
            + " EJ",
        ]

        p = wedges[0]
        text = texts[0]

        ang = (p.theta2 - p.theta1) / 2.0 + p.theta1
        y = np.sin(np.deg2rad(ang))
        x = np.cos(np.deg2rad(ang))
        horizontalalignment = {-1: "right", 1: "left"}[int(np.sign(x))]
        connectionstyle = "angle,angleA=0,angleB={}".format(ang)
        kw["arrowprops"].update({"connectionstyle": connectionstyle})
        self.ax.annotate(
            text,
            xy=(x, y),
            xytext=(1.2 * np.sign(x), 1.1 * y),
            horizontalalignment=horizontalalignment,
            **kw,
        )

        p = wedges[1]
        text = texts[1]

        ang = (p.theta2 - p.theta1) / 2.0 + p.theta1 - 45
        y = np.sin(np.deg2rad(ang)) * 0.7
        x = np.cos(np.deg2rad(ang)) * 0.00001
        horizontalalignment = {-1: "right", 1: "left"}[int(np.sign(x))]
        connectionstyle = "angle,angleA=0,angleB={}".format(ang)
        kw["arrowprops"].update({"connectionstyle": connectionstyle, "facecolor": "black"})
        self.ax.annotate(
            text,
            xy=(-x, y),
            xytext=(1.2 * np.sign(x), 2.0 * y),
            horizontalalignment=horizontalalignment,
            **kw,
        )

        legend_elements = [
            Line2D(
                [0],
                [0],
                color=color,
                lw=6,
                label="Electricity consumption of aviation in 2050",
            ),
            Line2D(
                [0],
                [0],
                color="grey",
                lw=6,
                label="Electricity resources allocated to aviation",
            ),
        ]

        self.ax.legend(handles=legend_elements, loc="center", bbox_to_anchor=[0.5, -0.12])

        self.fig.canvas.header_visible = False
        self.fig.canvas.toolbar_position = "bottom"
        # self.fig.canvas.layout.width = "auto"
        # self.fig.canvas.layout.height = "auto"
        # self.fig.set_tight_layout(True)

        self.fig.tight_layout()

    def update(self, data):
        self.parameters = data["float_inputs"]
        self.df = data["vector_outputs"]
        self.float_outputs = data["float_outputs"]
        self.years = data["years"]["full_years"]
        self.historic_years = data["years"]["historic_years"]
        self.prospective_years = data["years"]["prospective_years"]

        self.ax.clear()

        aviation_available_electricity = float(self.float_outputs["aviation_available_electricity"])
        electricity_consumption_end_year = float(
            self.float_outputs["electricity_consumption_end_year"]
        )
        available_electricity = float(self.parameters["available_electricity"])
        aviation_electricity_allocated_share = float(
            self.parameters["aviation_electricity_allocated_share"]
        )

        # Plot
        size = 0.3
        outer_vals = [
            electricity_consumption_end_year,
            available_electricity - electricity_consumption_end_year,
        ]
        inner_vals = [
            aviation_available_electricity,
            available_electricity - aviation_available_electricity,
        ]

        color = "skyblue"

        if electricity_consumption_end_year <= aviation_available_electricity:
            color = "green"
        outer_colors = [color, "white"]
        inner_colors = ["grey", "white"]

        self.wedges_out, _ = self.ax.pie(
            outer_vals,
            radius=1,
            colors=outer_colors,
            startangle=90,
            wedgeprops=dict(width=size, edgecolor="k", alpha=0.8),
        )

        self.wedges_in, _ = self.ax.pie(
            inner_vals,
            radius=1 - size,
            colors=inner_colors,
            startangle=90,
            wedgeprops=dict(width=size, edgecolor="k", alpha=0.8),
        )

        bbox_props = dict(boxstyle="square,pad=0.3", fc="w", ec="k", lw=0.72)
        kw = dict(arrowprops=dict(arrowstyle="-"), bbox=bbox_props, zorder=0, va="center")

        self.ax.set(
            aspect="equal",
            title="Estimated electricity consumption of aviation in 2050\nPresented as a % of the available elecltricity\n(AE) in 2050",
        )

        wedges = [self.wedges_out[0], self.wedges_in[0]]
        texts = [
            str(round(electricity_consumption_end_year, 1))
            + " EJ\ncorresponding to\n "
            + str(round(electricity_consumption_end_year / available_electricity * 100, 1))
            + "% of AE",
            str(round(aviation_electricity_allocated_share, 1))
            + "% of AE\ni.e.\n"
            + str(round(aviation_available_electricity, 1))
            + " EJ",
        ]

        p = wedges[0]
        text = texts[0]

        ang = (p.theta2 - p.theta1) / 2.0 + p.theta1
        y = np.sin(np.deg2rad(ang))
        x = np.cos(np.deg2rad(ang))
        horizontalalignment = {-1: "right", 1: "left"}[int(np.sign(x))]
        connectionstyle = "angle,angleA=0,angleB={}".format(ang)
        kw["arrowprops"].update({"connectionstyle": connectionstyle})
        self.ax.annotate(
            text,
            xy=(x, y),
            xytext=(1.2 * np.sign(x), 1.1 * y),
            horizontalalignment=horizontalalignment,
            **kw,
        )

        p = wedges[1]
        text = texts[1]

        ang = (p.theta2 - p.theta1) / 2.0 + p.theta1 - 45
        y = np.sin(np.deg2rad(ang)) * 0.7
        x = np.cos(np.deg2rad(ang)) * 0.00001
        horizontalalignment = {-1: "right", 1: "left"}[int(np.sign(x))]
        connectionstyle = "angle,angleA=0,angleB={}".format(ang)
        kw["arrowprops"].update({"connectionstyle": connectionstyle, "facecolor": "black"})
        self.ax.annotate(
            text,
            xy=(-x, y),
            xytext=(1.2 * np.sign(x), 2.0 * y),
            horizontalalignment=horizontalalignment,
            **kw,
        )

        legend_elements = [
            Line2D(
                [0],
                [0],
                color=color,
                lw=6,
                label="Electricity consumption of aviation in 2050",
            ),
            Line2D(
                [0],
                [0],
                color="grey",
                lw=6,
                label="Electricity resources allocated to aviation",
            ),
        ]

        self.ax.legend(handles=legend_elements, loc="center", bbox_to_anchor=[0.5, -0.12])

        self.fig.canvas.draw()


class MultidisciplinaryAssessmentPlot:
    def __init__(self, data):
        self.parameters = data["float_inputs"]
        self.df = data["vector_outputs"]
        self.float_outputs = data["float_outputs"]
        self.years = data["years"]["full_years"]
        self.historic_years = data["years"]["historic_years"]
        self.prospective_years = data["years"]["prospective_years"]

        self.fig, self.ax = plt.subplots(
            figsize=(plot_2_x, plot_2_y), subplot_kw={"projection": "polar"}
        )
        self.create_plot()

    def create_plot(self):
        # Carbon budget
        gross_carbon_budget = float(self.float_outputs["gross_carbon_budget_2050"])
        aviation_carbon_budget = float(self.float_outputs["aviation_carbon_budget"])
        cumulative_co2_emissions = float(self.df.loc[2050, "cumulative_co2_emissions"])

        # Biomass
        available_biomass_total = self.float_outputs["available_biomass_total"]
        aviation_available_biomass = self.float_outputs["aviation_available_biomass"]
        biomass_consumption_end_year = self.float_outputs["biomass_consumption_end_year"]

        # Electricity
        available_electricity_total = self.parameters["available_electricity"]
        aviation_available_electricity = self.float_outputs["aviation_available_electricity"]
        electricity_consumption_end_year = self.float_outputs["electricity_consumption_end_year"]

        # Effective radiative forcing
        equivalent_gross_carbon_budget = float(
            self.float_outputs["equivalent_gross_carbon_budget_2050"]
        )
        aviation_equivalent_carbon_budget = float(
            self.float_outputs["aviation_equivalent_carbon_budget"]
        )
        cumulative_total_equivalent_emissions = float(
            self.df.loc[2050, "cumulative_total_equivalent_emissions"]
        )

        categories = [
            "Climate\n(Total)",
            "Climate\n(CO2)",
            "Biomass",
            "Electricity",
        ]

        budgets = [
            aviation_equivalent_carbon_budget / equivalent_gross_carbon_budget * 100,
            aviation_carbon_budget / gross_carbon_budget * 100,
            aviation_available_biomass / available_biomass_total * 100,
            aviation_available_electricity / available_electricity_total * 100,
        ]

        consumptions = [
            np.max(
                [cumulative_total_equivalent_emissions / equivalent_gross_carbon_budget * 100, 0]
            ),
            cumulative_co2_emissions / gross_carbon_budget * 100,
            biomass_consumption_end_year / available_biomass_total * 100,
            electricity_consumption_end_year / available_electricity_total * 100,
        ]

        df_plot = pd.DataFrame(
            list(zip(categories, consumptions, budgets)),
            columns=["Category", "Consumption share", "Budget share"],
        )

        lowerLimit = 0

        # Compute max and min in the dataset
        max = df_plot[["Consumption share", "Budget share"]].max().max() + lowerLimit

        # Let's compute heights: they are a conversion of each item value in those new coordinates
        # In our example, 0 in the dataset will be converted to the lowerLimit (10)
        # The maximum will be converted to the upperLimit (100)
        slope = (max - lowerLimit) / max
        heights_consumption = slope * df_plot["Consumption share"] + lowerLimit
        heights_budget = slope * df_plot["Budget share"] + lowerLimit

        # # Compute the width of each bar. In total we have 2*Pi = 360°
        width = 2 * np.pi / len(df_plot.index) - 0.2
        #
        # # Compute the angle each bar is centered on:
        angles = np.linspace(np.pi / 4, (2 + 1 / 4) * np.pi, len(df_plot), endpoint=False)
        # Draw bars

        self.multidisciplinary_bars_plot = self.ax.bar(
            x=angles,
            height=heights_consumption,
            width=width,
            bottom=lowerLimit,
            linewidth=0,
            color="orange",
            edgecolor="none",
            alpha=0.4,
            label="Impacts",
        )

        self.multidisciplinary_bars_plot = self.ax.bar(
            x=angles,
            height=heights_consumption,
            width=width,
            bottom=lowerLimit,
            linewidth=2,
            color="none",
            edgecolor="orange",
            alpha=1,
        )

        self.multidisciplinary_bars_plot = self.ax.bar(
            x=angles,
            height=heights_budget,
            width=width,
            bottom=lowerLimit,
            linewidth=0,
            color="green",
            edgecolor="none",
            alpha=0.4,
            label="Budgets",
        )

        self.multidisciplinary_bars_plot = self.ax.bar(
            x=angles,
            height=heights_budget,
            width=width,
            bottom=lowerLimit,
            linewidth=2,
            color="none",
            edgecolor="green",
            alpha=1,
        )

        self.ax.set_xticks(angles)
        self.ax.set_xticklabels(df_plot["Category"], size=10, zorder=11)

        def percentage_formatter(x, pos):
            return f"{x:.0f}%"

        self.ax.set_rlabel_position(90)

        # Apply the custom tick formatter to the radial axis
        self.ax.yaxis.set_major_formatter(mticker.FuncFormatter(percentage_formatter))

        self.ax.grid(axis="x")
        self.ax.set_title(
            "Scenario impacts and allocated budgets\nPresented as a % of world budgets", y=1.05
        )

        for tick in self.ax.xaxis.get_major_ticks():
            tick.set_pad(10)

        self.ax.legend(loc="lower center", bbox_to_anchor=[0.5, -0.15], ncol=2)

        self.fig.canvas.header_visible = False
        self.fig.canvas.toolbar_position = "bottom"
        # self.fig.canvas.layout.width = "auto"
        # self.fig.canvas.layout.height = "auto"
        # Set the labels
        # self.fig.set_tight_layout(True)

        self.fig.tight_layout()

    def update(self, data):
        self.parameters = data["float_inputs"]
        self.df = data["vector_outputs"]
        self.float_outputs = data["float_outputs"]
        self.years = data["years"]["full_years"]
        self.historic_years = data["years"]["historic_years"]
        self.prospective_years = data["years"]["prospective_years"]

        self.ax.clear()

        gross_carbon_budget = float(self.float_outputs["gross_carbon_budget_2050"])
        aviation_carbon_budget = float(self.float_outputs["aviation_carbon_budget"])
        cumulative_co2_emissions = float(self.df.loc[2050, "cumulative_co2_emissions"])

        # Biomass
        available_biomass_total = self.float_outputs["available_biomass_total"]
        aviation_available_biomass = self.float_outputs["aviation_available_biomass"]
        biomass_consumption_end_year = self.float_outputs["biomass_consumption_end_year"]

        # Electricity
        available_electricity_total = self.parameters["available_electricity"]
        aviation_available_electricity = self.float_outputs["aviation_available_electricity"]
        electricity_consumption_end_year = self.float_outputs["electricity_consumption_end_year"]

        # Effective radiative forcing
        equivalent_gross_carbon_budget = float(
            self.float_outputs["equivalent_gross_carbon_budget_2050"]
        )
        aviation_equivalent_carbon_budget = float(
            self.float_outputs["aviation_equivalent_carbon_budget"]
        )
        cumulative_total_equivalent_emissions = float(
            self.df.loc[2050, "cumulative_total_equivalent_emissions"]
        )

        categories = [
            "Climate\n(Total)",
            "Climate\n(CO2)",
            "Biomass",
            "Electricity",
        ]

        budgets = [
            aviation_equivalent_carbon_budget / equivalent_gross_carbon_budget * 100,
            aviation_carbon_budget / gross_carbon_budget * 100,
            aviation_available_biomass / available_biomass_total * 100,
            aviation_available_electricity / available_electricity_total * 100,
        ]

        consumptions = [
            np.max(
                [cumulative_total_equivalent_emissions / equivalent_gross_carbon_budget * 100, 0]
            ),
            cumulative_co2_emissions / gross_carbon_budget * 100,
            biomass_consumption_end_year / available_biomass_total * 100,
            electricity_consumption_end_year / available_electricity_total * 100,
        ]

        df_plot = pd.DataFrame(
            list(zip(categories, consumptions, budgets)),
            columns=["Category", "Consumption share", "Budget share"],
        )

        lowerLimit = 0

        # Compute max and min in the dataset
        max = df_plot[["Consumption share", "Budget share"]].max().max() + lowerLimit

        # Let's compute heights: they are a conversion of each item value in those new coordinates
        # In our example, 0 in the dataset will be converted to the lowerLimit (10)
        # The maximum will be converted to the upperLimit (100)
        slope = (max - lowerLimit) / max
        heights_consumption = slope * df_plot["Consumption share"] + lowerLimit
        heights_budget = slope * df_plot["Budget share"] + lowerLimit

        # # Compute the width of each bar. In total we have 2*Pi = 360°
        width = 2 * np.pi / len(df_plot.index) - 0.2
        #
        # # Compute the angle each bar is centered on:
        angles = np.linspace(np.pi / 4, (2 + 1 / 4) * np.pi, len(df_plot), endpoint=False)
        # Draw bars

        self.multidisciplinary_bars_plot = self.ax.bar(
            x=angles,
            height=heights_consumption,
            width=width,
            bottom=lowerLimit,
            linewidth=0,
            color="orange",
            edgecolor="none",
            alpha=0.4,
            label="Impacts",
        )

        self.multidisciplinary_bars_plot = self.ax.bar(
            x=angles,
            height=heights_consumption,
            width=width,
            bottom=lowerLimit,
            linewidth=2,
            color="none",
            edgecolor="orange",
            alpha=1,
        )

        self.multidisciplinary_bars_plot = self.ax.bar(
            x=angles,
            height=heights_budget,
            width=width,
            bottom=lowerLimit,
            linewidth=0,
            color="green",
            edgecolor="none",
            alpha=0.4,
            label="Budgets",
        )

        self.multidisciplinary_bars_plot = self.ax.bar(
            x=angles,
            height=heights_budget,
            width=width,
            bottom=lowerLimit,
            linewidth=2,
            color="none",
            edgecolor="green",
            alpha=1,
        )

        self.ax.set_xticks(angles)
        self.ax.set_xticklabels(df_plot["Category"], size=10, zorder=11)

        def percentage_formatter(x, pos):
            return f"{x:.0f}%"

        self.ax.set_rlabel_position(90)

        # Apply the custom tick formatter to the radial axis
        self.ax.yaxis.set_major_formatter(mticker.FuncFormatter(percentage_formatter))

        self.ax.grid(axis="x")
        self.ax.set_title(
            "Scenario impacts and allocated budgets\nPresented as a % of world budgets", y=1.05
        )

        for tick in self.ax.xaxis.get_major_ticks():
            tick.set_pad(10)

        self.ax.legend(loc="lower center", bbox_to_anchor=[0.5, -0.15], ncol=2)

        self.fig.canvas.draw()



from typing import Tuple

import numpy as np
import pandas as pd

from aeromaps.models.base import AeromapsModel


class LoadFactor(AeromapsModel):
    def __init__(self, name="load_factor", *args, **kwargs):
        super().__init__(name=name, *args, **kwargs)

    def compute(
        self,
        load_factor_end_year: float = 0.0,
        covid_load_factor_2020: float = 0.0,
        rpk: pd.Series = pd.Series(dtype="float64"),
        ask_init: pd.Series = pd.Series(dtype="float64"),
    ) -> pd.Series:
        """Load factor calculation."""

        for k in range(self.historic_start_year, self.prospection_start_year):
            self.df.loc[k, "load_factor"] = rpk.loc[k] / ask_init.loc[k] * 100

        # Initialization for load factor
        load_factor_2019 = self.df.loc[2019, "load_factor"]

        # Parameters for the model
        a, b = self.parameters_load_factor_model(load_factor_2019, load_factor_end_year)

        for k in range(self.prospection_start_year, self.end_year + 1):
            self.df.loc[k, "load_factor"] = a * (k - 2019) ** 2 + b * (k - 2019) + load_factor_2019

        # Covid-19 : à refaire proprement
        self.df.loc[2020, "load_factor"] = covid_load_factor_2020

        load_factor = self.df["load_factor"]

        return load_factor

    @staticmethod
    def parameters_load_factor_model(load_factor_2019, load_factor_end_year):
        load_factor_2019 = load_factor_2019
        load_factor_2050 = load_factor_end_year
        # Calculate via derivative : 2ax+b
        derivative = 2 * (-5.62003082e-05) * 31 + 3.59670410e-03
        a = -(load_factor_2050 - load_factor_2019 - derivative * (2050 - 2019)) / (2050 - 2019) ** 2
        b = derivative - 2 * a * (2050 - 2019)
        return [a, b]



from typing import Tuple
import numpy as np
import pandas as pd

from aeromaps.models.base import AeromapsModel


class ERF(AeromapsModel):
    def __init__(self, name="effective_radiative_forcing", *args, **kwargs):
        super().__init__(name, *args, **kwargs)

    def compute(
        self,
        soot_emissions: pd.Series = pd.Series(dtype="float64"),
        co2_emissions: pd.Series = pd.Series(dtype="float64"),
        h2o_emissions: pd.Series = pd.Series(dtype="float64"),
        nox_emissions: pd.Series = pd.Series(dtype="float64"),
        sulfur_emissions: pd.Series = pd.Series(dtype="float64"),
        direct_co2_erf_2018_reference: float = 0.0,
        kerosene_emission_factor_2019: float = 0.0,
        erf_coefficient_soot: float = 0.0,
        erf_coefficient_co2: float = 0.0,
        erf_coefficient_contrails: float = 0.0,
        erf_coefficient_h2o: float = 0.0,
        erf_coefficient_nox: float = 0.0,
        erf_coefficient_sulfur: float = 0.0,
        total_aircraft_distance: pd.Series = pd.Series(dtype="float64"),
        operations_contrails_gain: pd.Series = pd.Series(dtype="float64"),
    ) -> Tuple[
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
    ]:
        """ERF calculation for the different climate impacts of aviation."""

        # CO2
        self.df["annual_co2_erf"] = co2_emissions * erf_coefficient_co2 / 1000

        # To improve
        reference_year = 2018
        co2_erf_2018_reference = (
            direct_co2_erf_2018_reference * kerosene_emission_factor_2019 / 73.2
        )

        self.df.loc[reference_year, "co2_erf"] = co2_erf_2018_reference
        for k in range(self.historic_start_year, reference_year):
            self.df.loc[self.historic_start_year + reference_year - 1 - k, "co2_erf"] = (
                self.df.loc[self.historic_start_year + reference_year - k, "co2_erf"]
                - self.df.loc[self.historic_start_year + reference_year - 1 - k, "annual_co2_erf"]
            )
        for k in range(reference_year + 1, self.end_year + 1):
            self.df.loc[k, "co2_erf"] = (
                self.df.loc[k - 1, "co2_erf"] + self.df.loc[k, "annual_co2_erf"]
            )

        annual_co2_erf = self.df["annual_co2_erf"]
        co2_erf = self.df["co2_erf"]

        # Contrails
        self.df["contrails_erf"] = (
            total_aircraft_distance
            * erf_coefficient_contrails
            * (1 - operations_contrails_gain / 100)
        )
        contrails_erf = self.df["contrails_erf"]

        # Others
        self.df["soot_erf"] = soot_emissions * erf_coefficient_soot
        self.df["h2o_erf"] = h2o_emissions * erf_coefficient_h2o
        self.df["nox_erf"] = nox_emissions * erf_coefficient_nox
        self.df["sulfur_erf"] = sulfur_emissions * erf_coefficient_sulfur
        soot_erf = self.df["soot_erf"]
        h2o_erf = self.df["h2o_erf"]
        nox_erf = self.df["nox_erf"]
        sulfur_erf = self.df["sulfur_erf"]

        # Total
        self.df["aerosol_erf"] = soot_erf + sulfur_erf
        self.df["total_erf"] = co2_erf + contrails_erf + h2o_erf + nox_erf + soot_erf + sulfur_erf
        aerosol_erf = self.df["aerosol_erf"]
        total_erf = self.df["total_erf"]

        return (
            annual_co2_erf,
            co2_erf,
            contrails_erf,
            soot_erf,
            h2o_erf,
            nox_erf,
            sulfur_erf,
            aerosol_erf,
            total_erf,
        )


class DetailedERF(AeromapsModel):
    def __init__(self, name="detailed_erf", *args, **kwargs):
        super().__init__(name, *args, **kwargs)

    def compute(
        self,
        co2_erf: pd.Series = pd.Series(dtype="float64"),
        contrails_erf: pd.Series = pd.Series(dtype="float64"),
        h2o_erf: pd.Series = pd.Series(dtype="float64"),
        nox_erf: pd.Series = pd.Series(dtype="float64"),
    ) -> Tuple[pd.Series, pd.Series, pd.Series]:
        """ERF calculation for helping plot display."""

        self.df["co2_h2o_erf"] = co2_erf + h2o_erf
        self.df["co2_h2o_nox_erf"] = co2_erf + h2o_erf + nox_erf
        self.df["co2_h2o_nox_contrails_erf"] = co2_erf + h2o_erf + nox_erf + contrails_erf

        co2_h2o_erf = self.df["co2_h2o_erf"]
        co2_h2o_nox_erf = self.df["co2_h2o_nox_erf"]
        co2_h2o_nox_contrails_erf = self.df["co2_h2o_nox_contrails_erf"]

        return co2_h2o_erf, co2_h2o_nox_erf, co2_h2o_nox_contrails_erf



"""A discipline interfacing a Python function."""
from __future__ import annotations

import ast
from inspect import getfullargspec
from inspect import getsource
from typing import Callable
from typing import Iterable
from typing import Sequence
from typing import Union
from typing import get_type_hints

import numpy as np
from numpy import array
from numpy import atleast_2d
from numpy import ndarray
import pandas as pd

from gemseo.core.data_processor import DataProcessor
from gemseo.core.discipline import MDODiscipline
from gemseo.utils.data_conversion import split_array_to_dict_of_arrays

from aeromaps.models.base import AeromapsModel

DataType = Union[float, ndarray]


class AutoPyDiscipline(MDODiscipline):
    """Wrap a Python function into a discipline.

    A simplified and straightforward way of integrating a discipline
    from a Python function.

    The Python function can take and return only numbers and NumPy arrays.

    The Python function may or may not include default values for input arguments,
    however, if the resulting :class:`.AutoPyDiscipline` is going to be placed inside
    an :class:`.MDF`, a :class:`.BiLevel` formulation or an :class:`.MDA`
    with strong couplings, then the Python function **must** assign default values for
    its input arguments.

    Example:
        >>> from gemseo.disciplines.auto_py import AutoPyDiscipline
        >>> from numpy import array
        >>> def my_function(x=0., y=0.):
        >>>     z1 = x + 2*y
        >>>     z2 = x + 2*y + 1
        >>>     return z1, z2
        >>>
        >>> discipline = AutoPyDiscipline(my_function)
        >>> discipline.execute()
        {'x': array([0.]), 'y': array([0.]), 'z1': array([0.]), 'z2': array([1.])}
        >>> discipline.execute({'x': array([1.]), 'y':array([-3.2])})
        {'x': array([1.]), 'y': array([-3.2]), 'z1': array([-5.4]), 'z2': array([-4.4])}
    """

    py_func: Callable[[DataType, ..., DataType], DataType]
    """The Python function to compute the outputs from the inputs."""

    use_arrays: bool
    """Whether the function is expected to take arrays as inputs and give outputs as
    arrays."""

    py_jac: Callable[[DataType, ..., DataType], ndarray] | None
    """The Python function to compute the Jacobian from the inputs."""

    input_names: list[str]
    """The names of the inputs."""

    output_names: list[str]
    """The names of the outputs."""

    data_processor: AutoDiscDataProcessor
    """A data processor forcing input data to float and output data to arrays."""

    sizes: dict[str, int]
    """The sizes of the input and output variables."""

    def __init__(
        self,
        py_func: Callable[[DataType, ..., DataType], DataType],
        py_jac: Callable[[DataType, ..., DataType], ndarray] | None = None,
        name: str | None = None,
        use_arrays: bool = False,
        grammar_type: MDODiscipline.GrammarType = MDODiscipline.GrammarType.JSON,
    ) -> None:
        """
        Args:
            py_func: The Python function to compute the outputs from the inputs.
            py_jac: The Python function to compute the Jacobian from the inputs;
                its output value must be a 2D NumPy array
                with rows correspond to the outputs
                and columns to the inputs.
            name: The name of the discipline. If ``None``, use the name of the Python
                function.
            use_arrays: Whether the function is expected
                to take arrays as inputs and give outputs as arrays.

        Raises:
            TypeError: When ``py_func`` is not callable.
        """  # noqa: D205 D212 D415
        if not callable(py_func):
            raise TypeError("py_func must be callable.")

        super().__init__(name=name or py_func.__name__, grammar_type=grammar_type)
        self.py_func = py_func
        self.use_arrays = use_arrays
        self.py_jac = py_jac
        self.input_names = getfullargspec(py_func)[0]
        if "self" in self.input_names:
            self.input_names.remove("self")
        self.input_grammar.update_from_names(self.input_names)
        self.output_names = []
        for node in ast.walk(ast.parse(getsource(py_func).strip())):
            if isinstance(node, ast.Return):
                value = node.value
                if isinstance(value, ast.Tuple):
                    temp_output_names = [elt.id for elt in value.elts]
                else:
                    temp_output_names = [value.id]

                if self.output_names and self.output_names != temp_output_names:
                    raise ValueError(
                        "Two return statements use different variable names; "
                        f"{self.output_names} and {temp_output_names}."
                    )
                else:
                    self.output_names = temp_output_names

        self.output_grammar.update_from_names(self.output_names)

        if not use_arrays:
            self.data_processor = AutoDiscDataProcessor(self.output_names)

        if self.py_jac is None:
            self.set_jacobian_approximation()

        self.default_inputs = to_arrays_dict(self._get_defaults())
        self.__sizes = {}
        self.__jac_shape = []
        self.__input_names_with_namespaces = []
        self.__output_names_with_namespaces = []

    def _get_defaults(self) -> dict[str, DataType]:
        """Return the default values of the input variables when available.

        The values are read from the signature of the Python function.

        Returns:
            The default values of the input variables.
        """
        args, _, _, defaults, _, _, _ = getfullargspec(self.py_func)
        if defaults is None:
            return {}

        n_defaults = len(defaults)
        names = args[-n_defaults:]
        return {names[i]: defaults[i] for i in range(n_defaults)}

    def _run(self) -> None:
        input_vals = self.get_input_data(with_namespaces=False)
        input_vals = self._convert_inputs(input_vals)
        output_values = self.py_func(**input_vals)
        if len(self.output_names) == 1:
            output_values = {self.output_names[0]: output_values}
        else:
            output_values = dict(zip(self.output_names, output_values))
        self.store_local_data(**output_values)

    def _compute_jacobian(
        self,
        inputs: Iterable[str] | None = None,
        outputs: Iterable[str] | None = None,
    ) -> None:
        """
        Raises:
            RuntimeError: When the analytic Jacobian :attr:`.py_jac` is ``None``.
            ValueError: When the Jacobian shape is inconsistent.
        """  # noqa: D205 D212 D415
        if self.py_jac is None:
            raise RuntimeError("The analytic Jacobian is missing.")

        if not self.__sizes:
            self.__sizes = {k: v.size for k, v in self.local_data.items()}

            in_to_ns = self.input_grammar.to_namespaced
            self.__input_names_with_namespaces = [
                in_to_ns[input_name] if input_name in in_to_ns else input_name
                for input_name in self.input_names
            ]
            out_to_ns = self.output_grammar.to_namespaced
            self.__output_names_with_namespaces = [
                out_to_ns[output_name] if output_name in out_to_ns else output_name
                for output_name in self.output_names
            ]
            self.__jac_shape = (
                sum(
                    self.__sizes[output_name] for output_name in self.__output_names_with_namespaces
                ),
                sum(self.__sizes[input_name] for input_name in self.__input_names_with_namespaces),
            )

        func_jac = self.py_jac(**self.get_input_data(with_namespaces=False))
        if len(func_jac.shape) < 2:
            func_jac = atleast_2d(func_jac)
        if func_jac.shape != self.__jac_shape:
            raise ValueError(
                f"The shape {func_jac.shape} "
                "of the Jacobian matrix "
                f"of the discipline {self.name} "
                f"provided by py_jac "
                "does not match "
                f"(output_size, input_size)={self.__jac_shape}."
            )

        self.jac = split_array_to_dict_of_arrays(
            func_jac,
            self.__sizes,
            self.__output_names_with_namespaces,
            self.__input_names_with_namespaces,
        )

    def _convert_inputs(self, input_vals):
        for name, val in input_vals.items():
            if issubclass(self._get_input_types()[name], pd.Series):
                input_vals[name] = pd.Series(val, index=self.model.df.index)
            elif issubclass(self._get_input_types()[name], int):
                input_vals[name] = int(val)
            else:
                pass
        return input_vals


class AutoDiscDataProcessor(DataProcessor):
    """A data processor forcing input data to float and output data to arrays.

    Convert all |g| scalar input data to floats, and convert all discipline output data
    to NumPy arrays.
    """

    out_names: Sequence[str]
    """The names of the outputs."""

    one_output: bool
    """Whether there is a single output."""

    def __init__(
        self,
        out_names: Sequence[str],
    ) -> None:
        """
        Args:
            out_names: The names of the outputs.
        """  # noqa: D205 D212 D415
        super().__init__()
        self.out_names = out_names
        self.one_output = len(out_names) == 1

    def pre_process_data(self, data: dict[str, DataType]) -> dict[str, DataType]:
        """Pre-process the input data.

        Execute a pre-processing of input data
        after they are checked by :meth:`~MDODiscipline.check_input_data`,
        and before the :meth:`~MDODiscipline._run` method of the discipline is called.

        Args:
            data: The data to be processed.

        Returns:
            The processed data
            where one-length NumPy arrays have been replaced with floats.
        """
        processed_data = data.copy()
        for key, val in data.items():
            if len(val) == 1:
                try:
                    processed_data[key] = float(val[0])
                except:
                    processed_data[key] = val[0]

        return processed_data

    def post_process_data(self, data: dict[str, DataType]) -> dict[str, ndarray]:
        """Post-process the output data.

        Execute a post-processing of the output data
        after the :meth:`~MDODiscipline._run` method of the discipline is called,
        and before they are checked by :meth:`~MDODiscipline.check_output_data`.

        Args:
            data: The data to be processed.

        Returns:
            The processed data with NumPy arrays as values.
        """
        processed_data = data.copy()
        for output_name, output_value in processed_data.items():
            if not isinstance(output_value, ndarray):
                processed_data[output_name] = array([output_value])

        return processed_data


def to_arrays_dict(data: dict[str, DataType]) -> dict[str, ndarray]:
    """Ensure that the values of a dictionary are NumPy arrays.

    Args:
        data: The dictionary whose values must be NumPy arrays.

    Returns:
        The dictionary with NumPy arrays as values.
    """
    for key, value in data.items():
        if not isinstance(value, ndarray):
            data[key] = array([value])
    return data


class AeromapsModelWrapper(AutoPyDiscipline):
    def __init__(self, model):
        self.model: AeromapsModel = model

        super(AeromapsModelWrapper, self).__init__(
            py_func=self.model.compute, grammar_type=MDODiscipline.GrammarType.SIMPLE
        )

        self.name = model.__class__.__name__

        self.update_defaults()

    def update_defaults(self):
        for input in self.get_input_data_names():
            # if self.model.parameters is None:
            #     self.default_inputs[input] = array([0])
            if hasattr(self.model.parameters, input):
                if issubclass(self._get_input_types()[input], pd.Series):
                    self.default_inputs[input] = array(getattr(self.model.parameters, input))
                else:
                    self.default_inputs[input] = array([getattr(self.model.parameters, input)])

    def _get_input_types(self):
        all_types = get_type_hints(self.py_func)
        input_types = {i: all_types[i] for i in all_types if i != "return"}
        return input_types

    def _get_output_types(self):
        all_types = get_type_hints(self.py_func)
        if "return" in all_types:
            types = all_types["return"]
            if hasattr(types, "__args__"):
                output_types_list = all_types["return"].__args__
            else:
                output_types_list = [types]
        else:
            output_types_list = []
        output_names = self.output_names

        output_types = {}
        for indx, output_type in enumerate(output_types_list):
            output_types[output_names[indx]] = output_type

        return output_types

    def update_elements(
        self,
        python_typing=False,  # type: bool
        **elements,  # type: Mapping[str,type]
    ):  # type: (...) -> None
        if python_typing:
            for element_name, element_value in elements.items():
                elements[element_name] = self.get_type_from_python_type(element_value)

        self._names_to_types.update(**elements)
        self._check_types()



from typing import Tuple

import numpy as np
import pandas as pd

from aeromaps.models.base import AeromapsModel


class EmissionsPerRPK(AeromapsModel):
    def __init__(self, name="emissions_per_rpk", *args, **kwargs):
        super().__init__(name=name, *args, **kwargs)

    def compute(
        self,
        co2_emissions_passenger: pd.Series = pd.Series(dtype="float64"),
        rpk: pd.Series = pd.Series(dtype="float64"),
    ) -> Tuple[pd.Series]:
        """CO2 emissions per Revenue Passenger Kilometer calculation."""

        self.df["co2_emissions_per_rpk"] = co2_emissions_passenger * 1e6 * 1e6 / rpk
        co2_emissions_per_rpk = self.df["co2_emissions_per_rpk"]

        return co2_emissions_per_rpk


class EmissionsPerRTK(AeromapsModel):
    def __init__(self, name="emissions_per_rpk", *args, **kwargs):
        super().__init__(name=name, *args, **kwargs)

    def compute(
        self,
        co2_emissions_freight: pd.Series = pd.Series(dtype="float64"),
        rtk: pd.Series = pd.Series(dtype="float64"),
    ) -> Tuple[pd.Series]:
        """CO2 emissions per Revenue Tonne Kilometer calculation."""

        self.df["co2_emissions_per_rtk"] = co2_emissions_freight * 1e6 * 1e6 / rtk
        co2_emissions_per_rtk = self.df["co2_emissions_per_rtk"]

        return co2_emissions_per_rtk


class DropinFuelConsumptionLiterPerPax100km(AeromapsModel):
    def __init__(self, name="dropin_fuel_consumption_liter_per_pax_100km", *args, **kwargs):
        super().__init__(name=name, *args, **kwargs)

    def compute(
        self,
        energy_consumption_passenger_kerosene: pd.Series = pd.Series(dtype="float64"),
        energy_consumption_passenger_biofuel: pd.Series = pd.Series(dtype="float64"),
        energy_consumption_passenger_electrofuel: pd.Series = pd.Series(dtype="float64"),
        lhv_kerosene: float = 0.0,
        lhv_biofuel: float = 0.0,
        lhv_electrofuel: float = 0.0,
        rpk: pd.Series = pd.Series(dtype="float64"),
    ) -> Tuple[pd.Series]:
        """Drop-in fuel consumption in liter per passenger per 100 km calculation."""

        density = 0.8
        self.df["dropin_fuel_consumption_liter_per_pax_100km"] = (
            (
                energy_consumption_passenger_kerosene / lhv_kerosene
                + energy_consumption_passenger_biofuel / lhv_biofuel
                + energy_consumption_passenger_electrofuel / lhv_electrofuel
            )
            / density
            * 100
            / rpk
        )
        dropin_fuel_consumption_liter_per_pax_100km = self.df[
            "dropin_fuel_consumption_liter_per_pax_100km"
        ]

        return dropin_fuel_consumption_liter_per_pax_100km



from typing import Tuple

import numpy as np
import pandas as pd
from scipy.interpolate import interp1d

from aeromaps.models.base import AeromapsModel


class BiofuelEmissionFactor(AeromapsModel):
    def __init__(self, name="biofuel_emission_factor", *args, **kwargs):
        super().__init__(name=name, *args, **kwargs)

    def compute(
        self,
        biofuel_hefa_fog_emission_factor_2020: float = 0.0,
        biofuel_hefa_fog_emission_factor_2050: float = 0.0,
        biofuel_hefa_others_emission_factor_2020: float = 0.0,
        biofuel_hefa_others_emission_factor_2050: float = 0.0,
        biofuel_ft_others_emission_factor_2020: float = 0.0,
        biofuel_ft_others_emission_factor_2050: float = 0.0,
        biofuel_ft_msw_emission_factor_2020: float = 0.0,
        biofuel_ft_msw_emission_factor_2050: float = 0.0,
        biofuel_atj_emission_factor_2020: float = 0.0,
        biofuel_atj_emission_factor_2050: float = 0.0,
        biofuel_hefa_fog_share: pd.Series = pd.Series(dtype="float64"),
        biofuel_hefa_others_share: pd.Series = pd.Series(dtype="float64"),
        biofuel_ft_others_share: pd.Series = pd.Series(dtype="float64"),
        biofuel_ft_msw_share: pd.Series = pd.Series(dtype="float64"),
        biofuel_atj_share: pd.Series = pd.Series(dtype="float64"),
    ) -> Tuple[pd.Series, pd.Series, pd.Series, pd.Series, pd.Series, pd.Series]:
        """Biofuel CO2 emission factor calculation using interpolation functions"""

        reference_years = [2020, self.end_year]

        # HEFA FOG
        reference_values_hefa_fog = [
            biofuel_hefa_fog_emission_factor_2020,
            biofuel_hefa_fog_emission_factor_2050,
        ]
        biofuel_hefa_fog_emission_factor_function = interp1d(
            reference_years, reference_values_hefa_fog, kind="linear"
        )
        for k in range(self.prospection_start_year, self.end_year + 1):
            self.df.loc[
                k, "biofuel_hefa_fog_emission_factor"
            ] = biofuel_hefa_fog_emission_factor_function(k)

        # HEFA OTHERS
        reference_values_hefa_others = [
            biofuel_hefa_others_emission_factor_2020,
            biofuel_hefa_others_emission_factor_2050,
        ]
        biofuel_hefa_others_emission_factor_function = interp1d(
            reference_years, reference_values_hefa_others, kind="linear"
        )
        for k in range(self.prospection_start_year, self.end_year + 1):
            self.df.loc[
                k, "biofuel_hefa_others_emission_factor"
            ] = biofuel_hefa_others_emission_factor_function(k)

        # FT OTHERS
        reference_values_ft_others = [
            biofuel_ft_others_emission_factor_2020,
            biofuel_ft_others_emission_factor_2050,
        ]
        biofuel_ft_others_emission_factor_function = interp1d(
            reference_years, reference_values_ft_others, kind="linear"
        )
        for k in range(self.prospection_start_year, self.end_year + 1):
            self.df.loc[
                k, "biofuel_ft_others_emission_factor"
            ] = biofuel_ft_others_emission_factor_function(k)

        # FT MSW
        reference_values_ft_msw = [
            biofuel_ft_msw_emission_factor_2020,
            biofuel_ft_msw_emission_factor_2050,
        ]
        biofuel_ft_msw_emission_factor_function = interp1d(
            reference_years, reference_values_ft_msw, kind="linear"
        )
        for k in range(self.prospection_start_year, self.end_year + 1):
            self.df.loc[
                k, "biofuel_ft_msw_emission_factor"
            ] = biofuel_ft_msw_emission_factor_function(k)

        # ATJ
        reference_values_atj = [
            biofuel_atj_emission_factor_2020,
            biofuel_atj_emission_factor_2050,
        ]
        biofuel_atj_emission_factor_function = interp1d(
            reference_years, reference_values_atj, kind="linear"
        )
        for k in range(self.prospection_start_year, self.end_year + 1):
            self.df.loc[k, "biofuel_atj_emission_factor"] = biofuel_atj_emission_factor_function(k)

        biofuel_hefa_fog_emission_factor = self.df["biofuel_hefa_fog_emission_factor"]
        biofuel_hefa_others_emission_factor = self.df["biofuel_hefa_others_emission_factor"]
        biofuel_ft_others_emission_factor = self.df["biofuel_ft_others_emission_factor"]
        biofuel_ft_msw_emission_factor = self.df["biofuel_ft_msw_emission_factor"]
        biofuel_atj_emission_factor = self.df["biofuel_atj_emission_factor"]

        # MEAN
        biofuel_mean_emission_factor = (
            biofuel_hefa_fog_emission_factor * biofuel_hefa_fog_share / 100
            + biofuel_hefa_others_emission_factor * biofuel_hefa_others_share / 100
            + biofuel_ft_others_emission_factor * biofuel_ft_others_share / 100
            + biofuel_ft_msw_emission_factor * biofuel_ft_msw_share / 100
            + biofuel_atj_emission_factor * biofuel_atj_share / 100
        )

        self.df.loc[:, "biofuel_mean_emission_factor"] = biofuel_mean_emission_factor

        return (
            biofuel_hefa_fog_emission_factor,
            biofuel_hefa_others_emission_factor,
            biofuel_ft_others_emission_factor,
            biofuel_ft_msw_emission_factor,
            biofuel_atj_emission_factor,
            biofuel_mean_emission_factor,
        )


class ElectricityEmissionFactor(AeromapsModel):
    def __init__(self, name="electricity_emission_factor", *args, **kwargs):
        super().__init__(name=name, *args, **kwargs)

    def compute(
        self,
        electricity_emission_factor_2020: float = 0.0,
        electricity_emission_factor_2030: float = 0.0,
        electricity_emission_factor_2040: float = 0.0,
        electricity_emission_factor_2050: float = 0.0,
    ) -> pd.Series:
        """Electricity CO2 emission factor calculation using interpolation function."""

        reference_years = [2020, 2030, 2040, self.parameters.end_year]
        reference_values = [
            electricity_emission_factor_2020,
            electricity_emission_factor_2030,
            electricity_emission_factor_2040,
            electricity_emission_factor_2050,
        ]
        electricity_emission_factor_function = interp1d(
            reference_years, reference_values, kind="quadratic"
        )
        for k in range(self.prospection_start_year, self.end_year + 1):
            self.df.loc[k, "electricity_emission_factor"] = electricity_emission_factor_function(k)
        electricity_emission_factor = self.df["electricity_emission_factor"]

        return electricity_emission_factor


class HydrogenEmissionFactor(AeromapsModel):
    def __init__(self, name="hydrogen_emission_factor", *args, **kwargs):
        super().__init__(name=name, *args, **kwargs)

    def compute(
        self,
        electricity_emission_factor: pd.Series = pd.Series(dtype="float64"),
        electrolysis_efficiency: pd.Series = pd.Series(dtype="float64"),
        liquefaction_efficiency: pd.Series = pd.Series(dtype="float64"),
        hydrogen_gas_ccs_emission_factor: float = 0.0,
        hydrogen_coal_ccs_emission_factor: float = 0.0,
        hydrogen_gas_emission_factor: float = 0.0,
        hydrogen_coal_emission_factor: float = 0.0,
        hydrogen_electrolysis_share: pd.Series = pd.Series(dtype="float64"),
        hydrogen_gas_ccs_share: pd.Series = pd.Series(dtype="float64"),
        hydrogen_coal_ccs_share: pd.Series = pd.Series(dtype="float64"),
        hydrogen_gas_share: pd.Series = pd.Series(dtype="float64"),
        hydrogen_coal_share: pd.Series = pd.Series(dtype="float64"),
    ) -> Tuple[pd.Series, pd.Series]:
        """Hydrogen CO2 emission factor calculation."""

        hydrogen_electrolysis_emission_factor = (
            electricity_emission_factor / 3.6 / electrolysis_efficiency / liquefaction_efficiency
        )

        hydrogen_mean_emission_factor = (
            hydrogen_electrolysis_emission_factor * hydrogen_electrolysis_share / 100
            + hydrogen_gas_ccs_emission_factor * hydrogen_gas_ccs_share / 100
            + hydrogen_coal_ccs_emission_factor * hydrogen_coal_ccs_share / 100
            + hydrogen_gas_emission_factor * hydrogen_gas_share / 100
            + hydrogen_coal_emission_factor * hydrogen_coal_share / 100
        )

        self.df.loc[
            :, "hydrogen_electrolysis_emission_factor"
        ] = hydrogen_electrolysis_emission_factor
        self.df.loc[:, "hydrogen_mean_emission_factor"] = hydrogen_mean_emission_factor

        return hydrogen_electrolysis_emission_factor, hydrogen_mean_emission_factor


class ElectrofuelEmissionFactor(AeromapsModel):
    def __init__(self, name="electrofuel_emission_factor", *args, **kwargs):
        super().__init__(name=name, *args, **kwargs)

    def compute(
        self,
        electricity_emission_factor: pd.Series = pd.Series(dtype="float64"),
        electrolysis_efficiency: pd.Series = pd.Series(dtype="float64"),
        electrofuel_hydrogen_efficiency: pd.Series = pd.Series(dtype="float64"),
    ) -> pd.Series:
        """Electrofuel CO2 emission factor calculation."""

        electrofuel_emission_factor = (
            electricity_emission_factor
            / 3.6
            / electrolysis_efficiency
            / electrofuel_hydrogen_efficiency
        )

        self.df.loc[:, "electrofuel_emission_factor"] = electrofuel_emission_factor

        return electrofuel_emission_factor


class KeroseneEmissionFactor(AeromapsModel):
    def __init__(self, name="kerosene_emission_factor", *args, **kwargs):
        super().__init__(name=name, *args, **kwargs)

    def compute(
        self,
        kerosene_emission_factor_2019: float = 0.0,
    ) -> pd.Series:
        """Kerosene CO2 emission factor calculation."""

        for k in range(self.historic_start_year, self.prospection_start_year):
            self.df.loc[k, "kerosene_emission_factor"] = kerosene_emission_factor_2019

        for k in range(self.prospection_start_year, self.end_year + 1):
            self.df.loc[k, "kerosene_emission_factor"] = kerosene_emission_factor_2019

        kerosene_emission_factor = self.df["kerosene_emission_factor"]

        return kerosene_emission_factor



from typing import Tuple

import numpy as np
import pandas as pd
from scipy.interpolate import interp1d

from aeromaps.models.base import AeromapsModel


class PassengerAircraftEfficiencySimple(AeromapsModel):
    def __init__(self, name="passenger_aircraft_efficiency_simple", *args, **kwargs):
        super().__init__(name=name, *args, **kwargs)

    def compute(
        self,
        energy_consumption_init: pd.Series = pd.Series(dtype="float64"),
        ask: pd.Series = pd.Series(dtype="float64"),
        short_range_energy_share_2019: float = 0.0,
        medium_range_energy_share_2019: float = 0.0,
        long_range_energy_share_2019: float = 0.0,
        short_range_rpk_share_2019: float = 0.0,
        medium_range_rpk_share_2019: float = 0.0,
        long_range_rpk_share_2019: float = 0.0,
        energy_per_ask_short_range_dropin_fuel_gain: float = 0.0,
        energy_per_ask_medium_range_dropin_fuel_gain: float = 0.0,
        energy_per_ask_long_range_dropin_fuel_gain: float = 0.0,
        relative_energy_per_ask_hydrogen_wrt_dropin_short_range: float = 0.0,
        relative_energy_per_ask_hydrogen_wrt_dropin_medium_range: float = 0.0,
        relative_energy_per_ask_hydrogen_wrt_dropin_long_range: float = 0.0,
        ask_short_range: pd.Series = pd.Series(dtype="float64"),
        ask_medium_range: pd.Series = pd.Series(dtype="float64"),
        ask_long_range: pd.Series = pd.Series(dtype="float64"),
        hydrogen_final_market_share_short_range: float = 0.0,
        hydrogen_introduction_year_short_range: float = 0.0,
        hydrogen_final_market_share_medium_range: float = 0.0,
        hydrogen_introduction_year_medium_range: float = 0.0,
        hydrogen_final_market_share_long_range: float = 0.0,
        hydrogen_introduction_year_long_range: float = 0.0,
        fleet_renewal_duration: float = 0.0,
        covid_energy_intensity_per_ask_increase_2020: float = 0.0,
    ) -> Tuple[
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
    ]:
        """Energy consumption per ASK (without operations) calculation using simple models."""

        # Initialization based on 2019 share
        energy_consumption_per_ask_init = energy_consumption_init / ask

        for k in range(self.historic_start_year, self.prospection_start_year):
            self.df.loc[k, "energy_per_ask_without_operations_short_range_dropin_fuel"] = (
                energy_consumption_per_ask_init.loc[k]
                * short_range_energy_share_2019
                / short_range_rpk_share_2019
            )
            self.df.loc[k, "energy_per_ask_without_operations_medium_range_dropin_fuel"] = (
                energy_consumption_per_ask_init.loc[k]
                * medium_range_energy_share_2019
                / medium_range_rpk_share_2019
            )
            self.df.loc[k, "energy_per_ask_without_operations_long_range_dropin_fuel"] = (
                energy_consumption_per_ask_init.loc[k]
                * long_range_energy_share_2019
                / long_range_rpk_share_2019
            )

        # Projections

        for k in range(self.prospection_start_year, self.end_year + 1):
            self.df.loc[
                k, "energy_per_ask_without_operations_short_range_dropin_fuel"
            ] = self.df.loc[k - 1, "energy_per_ask_without_operations_short_range_dropin_fuel"] * (
                1 - energy_per_ask_short_range_dropin_fuel_gain / 100
            )
            self.df.loc[
                k, "energy_per_ask_without_operations_medium_range_dropin_fuel"
            ] = self.df.loc[k - 1, "energy_per_ask_without_operations_medium_range_dropin_fuel"] * (
                1 - energy_per_ask_medium_range_dropin_fuel_gain / 100
            )
            self.df.loc[
                k, "energy_per_ask_without_operations_long_range_dropin_fuel"
            ] = self.df.loc[k - 1, "energy_per_ask_without_operations_long_range_dropin_fuel"] * (
                1 - energy_per_ask_long_range_dropin_fuel_gain / 100
            )

        self.df.loc[
            2020, "energy_per_ask_without_operations_short_range_dropin_fuel"
        ] = self.df.loc[2019, "energy_per_ask_without_operations_short_range_dropin_fuel"] * (
            1 + covid_energy_intensity_per_ask_increase_2020 / 100
        )
        self.df.loc[
            2020, "energy_per_ask_without_operations_medium_range_dropin_fuel"
        ] = self.df.loc[2019, "energy_per_ask_without_operations_medium_range_dropin_fuel"] * (
            1 + covid_energy_intensity_per_ask_increase_2020 / 100
        )
        self.df.loc[2020, "energy_per_ask_without_operations_long_range_dropin_fuel"] = self.df.loc[
            2019, "energy_per_ask_without_operations_long_range_dropin_fuel"
        ] * (1 + covid_energy_intensity_per_ask_increase_2020 / 100)

        energy_per_ask_without_operations_short_range_dropin_fuel = self.df[
            "energy_per_ask_without_operations_short_range_dropin_fuel"
        ]
        energy_per_ask_without_operations_medium_range_dropin_fuel = self.df[
            "energy_per_ask_without_operations_medium_range_dropin_fuel"
        ]
        energy_per_ask_without_operations_long_range_dropin_fuel = self.df[
            "energy_per_ask_without_operations_long_range_dropin_fuel"
        ]

        energy_per_ask_without_operations_short_range_hydrogen = (
            energy_per_ask_without_operations_short_range_dropin_fuel
            * relative_energy_per_ask_hydrogen_wrt_dropin_short_range
        )
        energy_per_ask_without_operations_medium_range_hydrogen = (
            energy_per_ask_without_operations_medium_range_dropin_fuel
            * relative_energy_per_ask_hydrogen_wrt_dropin_medium_range
        )
        energy_per_ask_without_operations_long_range_hydrogen = (
            energy_per_ask_without_operations_long_range_dropin_fuel
            * relative_energy_per_ask_hydrogen_wrt_dropin_long_range
        )

        self.df.loc[
            :, "energy_per_ask_without_operations_short_range_hydrogen"
        ] = energy_per_ask_without_operations_short_range_hydrogen
        self.df.loc[
            :, "energy_per_ask_without_operations_medium_range_hydrogen"
        ] = energy_per_ask_without_operations_medium_range_hydrogen
        self.df.loc[
            :, "energy_per_ask_without_operations_long_range_hydrogen"
        ] = energy_per_ask_without_operations_long_range_hydrogen

        # ASK Hydrogen

        ## Short range
        transition_year = hydrogen_introduction_year_short_range + fleet_renewal_duration / 2
        hydrogen_share_limit = 0.02 * hydrogen_final_market_share_short_range
        hydrogen_share_parameter = np.log(100 / 2 - 1) / (fleet_renewal_duration / 2)
        for k in range(self.historic_start_year, self.end_year + 1):
            if (
                hydrogen_final_market_share_short_range
                / (1 + np.exp(-hydrogen_share_parameter * (k - transition_year)))
                < hydrogen_share_limit
            ):
                self.df.loc[k, "ask_short_range_hydrogen_share"] = 0
            else:
                self.df.loc[
                    k, "ask_short_range_hydrogen_share"
                ] = hydrogen_final_market_share_short_range / (
                    1 + np.exp(-hydrogen_share_parameter * (k - transition_year))
                )

        ## Medium range
        transition_year = hydrogen_introduction_year_medium_range + fleet_renewal_duration / 2
        hydrogen_share_limit = 0.02 * hydrogen_final_market_share_medium_range
        hydrogen_share_parameter = np.log(100 / 2 - 1) / (fleet_renewal_duration / 2)
        for k in range(self.historic_start_year, self.end_year + 1):
            if (
                hydrogen_final_market_share_medium_range
                / (1 + np.exp(-hydrogen_share_parameter * (k - transition_year)))
                < hydrogen_share_limit
            ):
                self.df.loc[k, "ask_medium_range_hydrogen_share"] = 0
            else:
                self.df.loc[
                    k, "ask_medium_range_hydrogen_share"
                ] = hydrogen_final_market_share_medium_range / (
                    1 + np.exp(-hydrogen_share_parameter * (k - transition_year))
                )

        ## Long range
        transition_year = hydrogen_introduction_year_long_range + fleet_renewal_duration / 2
        hydrogen_share_limit = 0.02 * hydrogen_final_market_share_long_range
        hydrogen_share_parameter = np.log(100 / 2 - 1) / (fleet_renewal_duration / 2)
        for k in range(self.historic_start_year, self.end_year + 1):
            if (
                hydrogen_final_market_share_long_range
                / (1 + np.exp(-hydrogen_share_parameter * (k - transition_year)))
                < hydrogen_share_limit
            ):
                self.df.loc[k, "ask_long_range_hydrogen_share"] = 0
            else:
                self.df.loc[
                    k, "ask_long_range_hydrogen_share"
                ] = hydrogen_final_market_share_long_range / (
                    1 + np.exp(-hydrogen_share_parameter * (k - transition_year))
                )

        ask_short_range_hydrogen_share = self.df["ask_short_range_hydrogen_share"]
        ask_medium_range_hydrogen_share = self.df["ask_medium_range_hydrogen_share"]
        ask_long_range_hydrogen_share = self.df["ask_long_range_hydrogen_share"]

        ask_short_range_hydrogen = ask_short_range * ask_short_range_hydrogen_share / 100
        ask_medium_range_hydrogen = ask_medium_range * ask_medium_range_hydrogen_share / 100
        ask_long_range_hydrogen = ask_long_range * ask_long_range_hydrogen_share / 100
        self.df.loc[:, "ask_short_range_hydrogen"] = ask_short_range_hydrogen
        self.df.loc[:, "ask_medium_range_hydrogen"] = ask_medium_range_hydrogen
        self.df.loc[:, "ask_long_range_hydrogen"] = ask_long_range_hydrogen

        # ASK Drop-in fuel

        ask_short_range_dropin_fuel_share = 100 - ask_short_range_hydrogen_share
        ask_medium_range_dropin_fuel_share = 100 - ask_medium_range_hydrogen_share
        ask_long_range_dropin_fuel_share = 100 - ask_long_range_hydrogen_share
        self.df.loc[:, "ask_short_range_dropin_fuel_share"] = ask_short_range_dropin_fuel_share
        self.df.loc[:, "ask_medium_range_dropin_fuel_share"] = ask_medium_range_dropin_fuel_share
        self.df.loc[:, "ask_long_range_dropin_fuel_share"] = ask_long_range_dropin_fuel_share

        ask_short_range_dropin_fuel = ask_short_range - ask_short_range_hydrogen
        ask_medium_range_dropin_fuel = ask_medium_range - ask_medium_range_hydrogen
        ask_long_range_dropin_fuel = ask_long_range - ask_long_range_hydrogen
        self.df.loc[:, "ask_short_range_dropin_fuel"] = ask_short_range_dropin_fuel
        self.df.loc[:, "ask_medium_range_dropin_fuel"] = ask_medium_range_dropin_fuel
        self.df.loc[:, "ask_long_range_dropin_fuel"] = ask_long_range_dropin_fuel

        return (
            energy_per_ask_without_operations_short_range_dropin_fuel,
            energy_per_ask_without_operations_medium_range_dropin_fuel,
            energy_per_ask_without_operations_long_range_dropin_fuel,
            energy_per_ask_without_operations_short_range_hydrogen,
            energy_per_ask_without_operations_medium_range_hydrogen,
            energy_per_ask_without_operations_long_range_hydrogen,
            ask_short_range_dropin_fuel_share,
            ask_medium_range_dropin_fuel_share,
            ask_long_range_dropin_fuel_share,
            ask_short_range_hydrogen_share,
            ask_medium_range_hydrogen_share,
            ask_long_range_hydrogen_share,
            ask_short_range_dropin_fuel,
            ask_medium_range_dropin_fuel,
            ask_long_range_dropin_fuel,
            ask_short_range_hydrogen,
            ask_medium_range_hydrogen,
            ask_long_range_hydrogen,
        )


class PassengerAircraftEfficiencyComplex(AeromapsModel):
    def __init__(
        self, name="passenger_aircraft_efficiency_complex", fleet_model=None, *args, **kwargs
    ):
        super().__init__(name=name, *args, **kwargs)
        self.fleet_model = fleet_model

    def compute(
        self,
        dummy_fleet_model_output: np.ndarray,
        energy_consumption_init: pd.Series = pd.Series(dtype="float64"),
        ask: pd.Series = pd.Series(dtype="float64"),
        short_range_energy_share_2019: float = 0.0,
        medium_range_energy_share_2019: float = 0.0,
        long_range_energy_share_2019: float = 0.0,
        short_range_rpk_share_2019: float = 0.0,
        medium_range_rpk_share_2019: float = 0.0,
        long_range_rpk_share_2019: float = 0.0,
        covid_energy_intensity_per_ask_increase_2020: float = 0.0,
        ask_short_range: pd.Series = pd.Series(dtype="float64"),
        ask_medium_range: pd.Series = pd.Series(dtype="float64"),
        ask_long_range: pd.Series = pd.Series(dtype="float64"),
    ) -> Tuple[
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
    ]:
        ask_short_range_dropin_fuel_share = self.fleet_model.df["Short Range:share:dropin_fuel"]
        ask_medium_range_dropin_fuel_share = self.fleet_model.df["Medium Range:share:dropin_fuel"]
        ask_long_range_dropin_fuel_share = self.fleet_model.df["Long Range:share:dropin_fuel"]
        ask_short_range_hydrogen_share = self.fleet_model.df["Short Range:share:hydrogen"]
        ask_medium_range_hydrogen_share = self.fleet_model.df["Medium Range:share:hydrogen"]
        ask_long_range_hydrogen_share = self.fleet_model.df["Long Range:share:hydrogen"]

        energy_per_ask_without_operations_short_range_dropin_fuel = self.fleet_model.df[
            "Short Range:energy_consumption:dropin_fuel"
        ]
        energy_per_ask_without_operations_medium_range_dropin_fuel = self.fleet_model.df[
            "Medium Range:energy_consumption:dropin_fuel"
        ]
        energy_per_ask_without_operations_long_range_dropin_fuel = self.fleet_model.df[
            "Long Range:energy_consumption:dropin_fuel"
        ]
        energy_per_ask_without_operations_short_range_hydrogen = self.fleet_model.df[
            "Short Range:energy_consumption:hydrogen"
        ]
        energy_per_ask_without_operations_medium_range_hydrogen = self.fleet_model.df[
            "Medium Range:energy_consumption:hydrogen"
        ]
        energy_per_ask_without_operations_long_range_hydrogen = self.fleet_model.df[
            "Long Range:energy_consumption:hydrogen"
        ]

        """Energy consumption per ASK (without operations) calculation using complex models."""

        # Drop-in - Initialization based on 2019 share - To check for consistency
        energy_consumption_per_ask_init = energy_consumption_init / ask

        for k in range(self.historic_start_year, self.prospection_start_year):
            self.df.loc[k, "energy_per_ask_without_operations_short_range_dropin_fuel"] = (
                energy_consumption_per_ask_init.loc[k]
                * short_range_energy_share_2019
                / short_range_rpk_share_2019
            )
            self.df.loc[k, "energy_per_ask_without_operations_medium_range_dropin_fuel"] = (
                energy_consumption_per_ask_init.loc[k]
                * medium_range_energy_share_2019
                / medium_range_rpk_share_2019
            )
            self.df.loc[k, "energy_per_ask_without_operations_long_range_dropin_fuel"] = (
                energy_consumption_per_ask_init.loc[k]
                * long_range_energy_share_2019
                / long_range_rpk_share_2019
            )
        # Share
        self.df["ask_short_range_dropin_fuel_share"] = 100.0
        self.df["ask_medium_range_dropin_fuel_share"] = 100.0
        self.df["ask_long_range_dropin_fuel_share"] = 100.0

        # Hydrogen initialization
        # Energy consumption
        self.df["energy_per_ask_without_operations_short_range_hydrogen"] = 0.0
        self.df["energy_per_ask_without_operations_medium_range_hydrogen"] = 0.0
        self.df["energy_per_ask_without_operations_long_range_hydrogen"] = 0.0

        # Share
        self.df["ask_short_range_hydrogen_share"] = 0.0
        self.df["ask_medium_range_hydrogen_share"] = 0.0
        self.df["ask_long_range_hydrogen_share"] = 0.0

        # Drop-in - Projections
        for k in range(self.prospection_start_year, self.end_year + 1):
            self.df.loc[
                k, "energy_per_ask_without_operations_short_range_dropin_fuel"
            ] = energy_per_ask_without_operations_short_range_dropin_fuel.loc[k]
            self.df.loc[
                k, "energy_per_ask_without_operations_medium_range_dropin_fuel"
            ] = energy_per_ask_without_operations_medium_range_dropin_fuel.loc[k]
            self.df.loc[
                k, "energy_per_ask_without_operations_long_range_dropin_fuel"
            ] = energy_per_ask_without_operations_long_range_dropin_fuel.loc[k]

        self.df.loc[
            2020, "energy_per_ask_without_operations_short_range_dropin_fuel"
        ] = self.df.loc[2019, "energy_per_ask_without_operations_short_range_dropin_fuel"] * (
            1 + covid_energy_intensity_per_ask_increase_2020 / 100
        )
        self.df.loc[
            2020, "energy_per_ask_without_operations_medium_range_dropin_fuel"
        ] = self.df.loc[2019, "energy_per_ask_without_operations_medium_range_dropin_fuel"] * (
            1 + covid_energy_intensity_per_ask_increase_2020 / 100
        )
        self.df.loc[2020, "energy_per_ask_without_operations_long_range_dropin_fuel"] = self.df.loc[
            2019, "energy_per_ask_without_operations_long_range_dropin_fuel"
        ] * (1 + covid_energy_intensity_per_ask_increase_2020 / 100)

        energy_per_ask_without_operations_short_range_dropin_fuel = self.df[
            "energy_per_ask_without_operations_short_range_dropin_fuel"
        ]
        energy_per_ask_without_operations_medium_range_dropin_fuel = self.df[
            "energy_per_ask_without_operations_medium_range_dropin_fuel"
        ]
        energy_per_ask_without_operations_long_range_dropin_fuel = self.df[
            "energy_per_ask_without_operations_long_range_dropin_fuel"
        ]

        # Hydrogen
        for k in range(self.prospection_start_year + 1, self.end_year + 1):
            self.df.loc[
                k, "energy_per_ask_without_operations_short_range_hydrogen"
            ] = energy_per_ask_without_operations_short_range_hydrogen.loc[k]
            self.df.loc[
                k, "energy_per_ask_without_operations_medium_range_hydrogen"
            ] = energy_per_ask_without_operations_medium_range_hydrogen.loc[k]
            self.df.loc[
                k, "energy_per_ask_without_operations_long_range_hydrogen"
            ] = energy_per_ask_without_operations_long_range_hydrogen.loc[k]

        energy_per_ask_without_operations_short_range_hydrogen = self.df[
            "energy_per_ask_without_operations_short_range_hydrogen"
        ]
        energy_per_ask_without_operations_medium_range_hydrogen = self.df[
            "energy_per_ask_without_operations_medium_range_hydrogen"
        ]
        energy_per_ask_without_operations_long_range_hydrogen = self.df[
            "energy_per_ask_without_operations_long_range_hydrogen"
        ]

        # Share
        self.df.loc[
            self.prospection_start_year : self.end_year + 1, "ask_short_range_dropin_fuel_share"
        ] = ask_short_range_dropin_fuel_share
        self.df.loc[
            self.prospection_start_year : self.end_year + 1, "ask_medium_range_dropin_fuel_share"
        ] = ask_medium_range_dropin_fuel_share
        self.df.loc[
            self.prospection_start_year : self.end_year + 1, "ask_long_range_dropin_fuel_share"
        ] = ask_long_range_dropin_fuel_share
        self.df.loc[
            self.prospection_start_year : self.end_year + 1, "ask_short_range_hydrogen_share"
        ] = ask_short_range_hydrogen_share
        self.df.loc[
            self.prospection_start_year : self.end_year + 1, "ask_medium_range_hydrogen_share"
        ] = ask_medium_range_hydrogen_share
        self.df.loc[
            self.prospection_start_year : self.end_year + 1, "ask_long_range_hydrogen_share"
        ] = ask_long_range_hydrogen_share

        # ASK
        ask_short_range_dropin_fuel = (
            ask_short_range * self.df["ask_short_range_dropin_fuel_share"] / 100
        )
        ask_medium_range_dropin_fuel = (
            ask_medium_range * self.df["ask_medium_range_dropin_fuel_share"] / 100
        )
        ask_long_range_dropin_fuel = (
            ask_long_range * self.df["ask_long_range_dropin_fuel_share"] / 100
        )
        ask_short_range_hydrogen = ask_short_range * self.df["ask_short_range_hydrogen_share"] / 100
        ask_medium_range_hydrogen = (
            ask_medium_range * self.df["ask_medium_range_hydrogen_share"] / 100
        )
        ask_long_range_hydrogen = ask_long_range * self.df["ask_long_range_hydrogen_share"] / 100

        self.df.loc[:, "ask_short_range_dropin_fuel"] = ask_short_range_dropin_fuel
        self.df.loc[:, "ask_medium_range_dropin_fuel"] = ask_medium_range_dropin_fuel
        self.df.loc[:, "ask_long_range_dropin_fuel"] = ask_long_range_dropin_fuel
        self.df.loc[:, "ask_short_range_hydrogen"] = ask_short_range_hydrogen
        self.df.loc[:, "ask_medium_range_hydrogen"] = ask_medium_range_hydrogen
        self.df.loc[:, "ask_long_range_hydrogen"] = ask_long_range_hydrogen

        return (
            energy_per_ask_without_operations_short_range_dropin_fuel,
            energy_per_ask_without_operations_medium_range_dropin_fuel,
            energy_per_ask_without_operations_long_range_dropin_fuel,
            energy_per_ask_without_operations_short_range_hydrogen,
            energy_per_ask_without_operations_medium_range_hydrogen,
            energy_per_ask_without_operations_long_range_hydrogen,
            ask_short_range_dropin_fuel_share,
            ask_medium_range_dropin_fuel_share,
            ask_long_range_dropin_fuel_share,
            ask_short_range_hydrogen_share,
            ask_medium_range_hydrogen_share,
            ask_long_range_hydrogen_share,
            ask_short_range_dropin_fuel,
            ask_medium_range_dropin_fuel,
            ask_long_range_dropin_fuel,
            ask_short_range_hydrogen,
            ask_medium_range_hydrogen,
            ask_long_range_hydrogen,
        )


class FreightAircraftEfficiency(AeromapsModel):
    def __init__(self, name="freight_aircraft_efficiency", *args, **kwargs):
        super().__init__(name=name, *args, **kwargs)

    def compute(
        self,
        energy_consumption_init: pd.Series = pd.Series(dtype="float64"),
        rtk: pd.Series = pd.Series(dtype="float64"),
        freight_energy_share_2019: float = 0.0,
        energy_per_ask_without_operations_short_range_dropin_fuel: pd.Series = pd.Series(
            dtype="float64"
        ),
        energy_per_ask_without_operations_medium_range_dropin_fuel: pd.Series = pd.Series(
            dtype="float64"
        ),
        energy_per_ask_without_operations_long_range_dropin_fuel: pd.Series = pd.Series(
            dtype="float64"
        ),
        energy_per_ask_without_operations_short_range_hydrogen: pd.Series = pd.Series(
            dtype="float64"
        ),
        energy_per_ask_without_operations_medium_range_hydrogen: pd.Series = pd.Series(
            dtype="float64"
        ),
        energy_per_ask_without_operations_long_range_hydrogen: pd.Series = pd.Series(
            dtype="float64"
        ),
        ask: pd.Series = pd.Series(dtype="float64"),
        ask_short_range: pd.Series = pd.Series(dtype="float64"),
        ask_medium_range: pd.Series = pd.Series(dtype="float64"),
        ask_long_range: pd.Series = pd.Series(dtype="float64"),
        ask_short_range_dropin_fuel: pd.Series = pd.Series(dtype="float64"),
        ask_medium_range_dropin_fuel: pd.Series = pd.Series(dtype="float64"),
        ask_long_range_dropin_fuel: pd.Series = pd.Series(dtype="float64"),
        ask_short_range_hydrogen_share: pd.Series = pd.Series(dtype="float64"),
        ask_medium_range_hydrogen_share: pd.Series = pd.Series(dtype="float64"),
        ask_long_range_hydrogen_share: pd.Series = pd.Series(dtype="float64"),
        covid_energy_intensity_per_ask_increase_2020: float = 0.0,
    ) -> Tuple[pd.Series, pd.Series, pd.Series, pd.Series, pd.Series, pd.Series]:
        """Energy consumption per RTK (without operations) calculation."""

        # Initialization based on 2019 share: to correct to include load factor
        for k in range(self.historic_start_year, self.prospection_start_year):
            self.df.loc[k, "energy_per_rtk_without_operations_freight_dropin_fuel"] = (
                energy_consumption_init.loc[k] / rtk.loc[k] * freight_energy_share_2019 / 100
            )

        # Projections
        energy_per_rtk_without_operations_freight_dropin_fuel_short_range_k = self.df.loc[
            2019, "energy_per_rtk_without_operations_freight_dropin_fuel"
        ]
        energy_per_rtk_without_operations_freight_dropin_fuel_medium_range_k = self.df.loc[
            2019, "energy_per_rtk_without_operations_freight_dropin_fuel"
        ]
        energy_per_rtk_without_operations_freight_dropin_fuel_long_range_k = self.df.loc[
            2019, "energy_per_rtk_without_operations_freight_dropin_fuel"
        ]
        for k in range(self.prospection_start_year, self.end_year + 1):
            energy_per_rtk_without_operations_freight_dropin_fuel_short_range_k = (
                energy_per_rtk_without_operations_freight_dropin_fuel_short_range_k
                * energy_per_ask_without_operations_short_range_dropin_fuel.loc[k]
                / energy_per_ask_without_operations_short_range_dropin_fuel.loc[k - 1]
            )
            energy_per_rtk_without_operations_freight_dropin_fuel_medium_range_k = (
                energy_per_rtk_without_operations_freight_dropin_fuel_medium_range_k
                * energy_per_ask_without_operations_medium_range_dropin_fuel.loc[k]
                / energy_per_ask_without_operations_medium_range_dropin_fuel.loc[k - 1]
            )
            energy_per_rtk_without_operations_freight_dropin_fuel_long_range_k = (
                energy_per_rtk_without_operations_freight_dropin_fuel_long_range_k
                * energy_per_ask_without_operations_long_range_dropin_fuel.loc[k]
                / energy_per_ask_without_operations_long_range_dropin_fuel.loc[k - 1]
            )
            self.df.loc[k, "energy_per_rtk_without_operations_freight_dropin_fuel"] = (
                energy_per_rtk_without_operations_freight_dropin_fuel_short_range_k
                * ask_short_range_dropin_fuel.loc[k]
                / (
                    ask_short_range_dropin_fuel.loc[k]
                    + ask_medium_range_dropin_fuel.loc[k]
                    + ask_long_range_dropin_fuel.loc[k]
                )
                + energy_per_rtk_without_operations_freight_dropin_fuel_medium_range_k
                * ask_medium_range_dropin_fuel.loc[k]
                / (
                    ask_short_range_dropin_fuel.loc[k]
                    + ask_medium_range_dropin_fuel.loc[k]
                    + ask_long_range_dropin_fuel.loc[k]
                )
                + energy_per_rtk_without_operations_freight_dropin_fuel_long_range_k
                * ask_long_range_dropin_fuel.loc[k]
                / (
                    ask_short_range_dropin_fuel.loc[k]
                    + ask_medium_range_dropin_fuel.loc[k]
                    + ask_long_range_dropin_fuel.loc[k]
                )
            )

        # Covid
        self.df.loc[2020, "energy_per_rtk_without_operations_freight_dropin_fuel"] = self.df.loc[
            2019, "energy_per_rtk_without_operations_freight_dropin_fuel"
        ] * (1 + covid_energy_intensity_per_ask_increase_2020 / 100)

        energy_per_rtk_without_operations_freight_dropin_fuel = self.df[
            "energy_per_rtk_without_operations_freight_dropin_fuel"
        ]

        rtk_hydrogen_share = (
            ask_short_range_hydrogen_share * (ask_short_range / ask)
            + ask_medium_range_hydrogen_share * (ask_medium_range / ask)
            + ask_long_range_hydrogen_share * (ask_long_range / ask)
        )
        rtk_dropin_fuel_share = 100 - rtk_hydrogen_share
        self.df.loc[:, "rtk_hydrogen_share"] = rtk_hydrogen_share
        self.df.loc[:, "rtk_dropin_fuel_share"] = rtk_dropin_fuel_share

        rtk_hydrogen = rtk * rtk_hydrogen_share / 100
        rtk_dropin_fuel = rtk * rtk_dropin_fuel_share / 100
        self.df.loc[:, "rtk_hydrogen"] = rtk_hydrogen
        self.df.loc[:, "rtk_dropin_fuel"] = rtk_dropin_fuel

        relative_energy_per_ask_hydrogen_wrt_dropin_short_range = (
            energy_per_ask_without_operations_short_range_hydrogen
            / energy_per_ask_without_operations_short_range_dropin_fuel
        )
        relative_energy_per_ask_hydrogen_wrt_dropin_medium_range = (
            energy_per_ask_without_operations_medium_range_hydrogen
            / energy_per_ask_without_operations_medium_range_dropin_fuel
        )
        relative_energy_per_ask_hydrogen_wrt_dropin_long_range = (
            energy_per_ask_without_operations_long_range_hydrogen
            / energy_per_ask_without_operations_long_range_dropin_fuel
        )

        for k in range(self.historic_start_year, self.end_year + 1):
            if rtk_hydrogen_share.loc[k] == 0:
                self.df.loc[k, "energy_per_rtk_without_operations_freight_hydrogen"] = self.df.loc[
                    k, "energy_per_rtk_without_operations_freight_dropin_fuel"
                ]
            else:
                self.df.loc[
                    k, "energy_per_rtk_without_operations_freight_hydrogen"
                ] = energy_per_rtk_without_operations_freight_dropin_fuel.loc[k] * (
                    relative_energy_per_ask_hydrogen_wrt_dropin_short_range.loc[k]
                    * ask_short_range_hydrogen_share.loc[k]
                    * (ask_short_range.loc[k] / ask.loc[k])
                    / rtk_hydrogen_share.loc[k]
                    + relative_energy_per_ask_hydrogen_wrt_dropin_medium_range.loc[k]
                    * ask_medium_range_hydrogen_share.loc[k]
                    * (ask_medium_range.loc[k] / ask.loc[k])
                    / rtk_hydrogen_share.loc[k]
                    + relative_energy_per_ask_hydrogen_wrt_dropin_long_range.loc[k]
                    * ask_long_range_hydrogen_share.loc[k]
                    * (ask_long_range.loc[k] / ask.loc[k])
                    / rtk_hydrogen_share.loc[k]
                )

        energy_per_rtk_without_operations_freight_hydrogen = self.df[
            "energy_per_rtk_without_operations_freight_hydrogen"
        ]

        return (
            energy_per_rtk_without_operations_freight_dropin_fuel,
            energy_per_rtk_without_operations_freight_hydrogen,
            rtk_dropin_fuel_share,
            rtk_hydrogen_share,
            rtk_dropin_fuel,
            rtk_hydrogen,
        )



from typing import Tuple

import numpy as np
import pandas as pd
from scipy.interpolate import interp1d

from aeromaps.models.base import AeromapsModel


class RTK(AeromapsModel):
    def __init__(self, name="rtk", *args, **kwargs):
        super().__init__(name=name, *args, **kwargs)

    def compute(
        self,
        rtk_init: pd.Series = pd.Series(dtype="float64"),
        covid_start_year: int = 0,
        covid_rpk_drop_start_year: float = 0.0,
        covid_end_year: int = 0,
        covid_end_year_reference_rpk_ratio: float = 0.0,
        growth_rate_2020_2030_freight: float = 0.0,
        growth_rate_2030_2040_freight: float = 0.0,
        growth_rate_2040_2050_freight: float = 0.0,
    ) -> Tuple[pd.Series, float, float,]:
        """RTK calculation."""

        # Initialization
        self.df.loc[self.prospection_start_year - 1, "rtk"] = rtk_init.loc[
            self.prospection_start_year - 1
        ]

        # Covid functions
        reference_years = [covid_start_year, covid_end_year]
        reference_values_covid = [
            1 - covid_rpk_drop_start_year / 100,
            covid_end_year_reference_rpk_ratio / 100,
        ]
        covid_function = interp1d(reference_years, reference_values_covid, kind="linear")

        # Main
        for k in range(covid_start_year, covid_end_year + 1):
            self.df.loc[k, "rtk"] = self.df.loc[covid_start_year - 1, "rtk"] * covid_function(k)
        for k in range(covid_end_year + 1, 2031):
            self.df.loc[k, "rtk"] = self.df.loc[k - 1, "rtk"] * (
                1 + growth_rate_2020_2030_freight / 100
            )
        for k in range(2031, 2041):
            self.df.loc[k, "rtk"] = self.df.loc[k - 1, "rtk"] * (
                1 + growth_rate_2030_2040_freight / 100
            )
        for k in range(2041, self.end_year + 1):
            self.df.loc[k, "rtk"] = self.df.loc[k - 1, "rtk"] * (
                1 + growth_rate_2040_2050_freight / 100
            )

        for k in range(self.historic_start_year, self.prospection_start_year):
            self.df.loc[k, "rtk"] = rtk_init.loc[k]
        rtk = self.df["rtk"]

        # Compound Annual Growth Rate (CAGR)
        cagr_rtk = 100 * (
            (
                self.df.loc[self.end_year, "rtk"]
                / self.df.loc[self.prospection_start_year - 1, "rtk"]
            )
            ** (1 / (self.end_year - self.prospection_start_year))
            - 1
        )

        # Prospective evolution of RTK (between prospection_start_year-1 and end_year)
        prospective_evolution_rtk = 100 * (
            self.df.loc[self.end_year, "rtk"] / self.df.loc[self.prospection_start_year - 1, "rtk"]
            - 1
        )

        self.float_outputs["cagr_rtk"] = cagr_rtk
        self.float_outputs["prospective_evolution_rtk"] = prospective_evolution_rtk

        return (
            rtk,
            cagr_rtk,
            prospective_evolution_rtk,
        )


class RTKReference(AeromapsModel):
    def __init__(self, name="rtk_reference", *args, **kwargs):
        super().__init__(name=name, *args, **kwargs)

    def compute(
        self,
        rtk: pd.Series = pd.Series(dtype="float64"),
        growth_rate_2020_2030_reference: float = 0.0,
        growth_rate_2030_2040_reference: float = 0.0,
        growth_rate_2040_2050_reference: float = 0.0,
        covid_start_year: int = 0,
        covid_rpk_drop_start_year: int = 0,
        covid_end_year: int = 0,
        covid_end_year_reference_rpk_ratio: int = 0,
    ) -> pd.Series:
        """RTK reference calculation."""

        for k in range(self.historic_start_year, self.prospection_start_year):
            self.df.loc[k, "rtk_reference"] = rtk.loc[k]

        covid_start_year = int(covid_start_year)
        covid_rpk_drop_start_year = int(covid_rpk_drop_start_year)
        covid_end_year = int(covid_end_year)
        covid_end_year_reference_rpk_ratio = int(covid_end_year_reference_rpk_ratio)

        self.df.loc[covid_start_year - 1, "rtk_reference"] = rtk.loc[covid_start_year - 1]

        # Covid functions
        reference_years = [covid_start_year, covid_end_year]
        reference_values_covid = [
            1 - covid_rpk_drop_start_year / 100,
            covid_end_year_reference_rpk_ratio / 100,
        ]
        covid_function = interp1d(reference_years, reference_values_covid, kind="linear")

        for k in range(covid_start_year, covid_end_year + 1):
            self.df.loc[k, "rtk_reference"] = self.df.loc[
                covid_start_year - 1, "rtk_reference"
            ] * covid_function(k)
        for k in range(covid_end_year + 1, 2031):
            self.df.loc[k, "rtk_reference"] = self.df.loc[k - 1, "rtk_reference"] * (
                1 + growth_rate_2020_2030_reference / 100
            )
        for k in range(2031, 2041):
            self.df.loc[k, "rtk_reference"] = self.df.loc[k - 1, "rtk_reference"] * (
                1 + growth_rate_2030_2040_reference / 100
            )
        for k in range(2041, self.end_year + 1):
            self.df.loc[k, "rtk_reference"] = self.df.loc[k - 1, "rtk_reference"] * (
                1 + growth_rate_2040_2050_reference / 100
            )

        rtk_reference = self.df["rtk_reference"]

        return rtk_reference



# @Time : 24/02/2023 17:44
# @Author : a.salgas
# @File : power_to_liquid.py
# @Software: PyCharm

from typing import Tuple, Union, Any

import numpy as np
import pandas as pd
from pandas import Series

from aeromaps.models.base import AeromapsModel

from scipy.interpolate import interp1d


class ElectrofuelCost(AeromapsModel):
    def __init__(self, name="electrofuel_cost", *args, **kwargs):
        super().__init__(name, *args, **kwargs)

    def compute(
        self,
        energy_consumption_electrofuel: pd.Series = pd.Series(dtype="float64"),
        electrofuel_eis_capex: pd.Series = pd.Series(dtype="float64"),
        electrofuel_eis_fixed_opex: pd.Series = pd.Series(dtype="float64"),
        electrofuel_eis_var_opex: pd.Series = pd.Series(dtype="float64"),
        electrolysis_efficiency: pd.Series = pd.Series(dtype="float64"),
        electrofuel_hydrogen_efficiency: pd.Series = pd.Series(dtype="float64"),
        kerosene_market_price: pd.Series = pd.Series(dtype="float64"),
        kerosene_emission_factor: pd.Series = pd.Series(dtype="float64"),
        electrofuel_emission_factor: pd.Series = pd.Series(dtype="float64"),
        electricity_market_price: pd.Series = pd.Series(dtype="float64"),
        co2_market_price: pd.Series = pd.Series(dtype="float64"),
        electrofuel_eis_specific_co2: pd.Series = pd.Series(dtype="float64"),
        electricity_load_factor: float = 0.0,
        carbon_tax: pd.Series = pd.Series(dtype="float64"),
    ) -> Tuple[
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
    ]:

        ######## HYDROGEN PRODUCTION ########

        #### ELECTROLYSIS ####

        (
            electrofuel_plant_building_scenario,
            electrofuel_plant_building_cost,
            electrofuel_total_cost,
            electrofuel_capex_cost,
            electrofuel_opex_cost,
            electrofuel_elec_cost,
            electrofuel_co2_cost,
        ) = self.electrofuel_computation(
            electrofuel_eis_capex,
            electrofuel_eis_fixed_opex,
            electrofuel_eis_var_opex,
            electrolysis_efficiency,
            electrofuel_hydrogen_efficiency,
            electricity_market_price,
            energy_consumption_electrofuel,
            electricity_load_factor,
            co2_market_price,
            electrofuel_eis_specific_co2,
        )

        self.df.loc[:, "electrofuel_plant_building_scenario"] = electrofuel_plant_building_scenario
        self.df.loc[:, "electrofuel_plant_building_cost"] = electrofuel_plant_building_cost
        self.df.loc[:, "electrofuel_total_cost"] = electrofuel_total_cost
        self.df.loc[:, "electrofuel_capex_cost"] = electrofuel_capex_cost
        self.df.loc[:, "electrofuel_opex_cost"] = electrofuel_opex_cost
        self.df.loc[:, "electrofuel_elec_cost"] = electrofuel_elec_cost
        self.df.loc[:, "electrofuel_co2_cost"] = electrofuel_co2_cost

        ######## SYNTHESIS ########

        fuel_lhv = 35.3  # MJ/L

        # Compute the total cost premium (M€)
        electrofuel_cost_premium = (
            electrofuel_total_cost
            - (energy_consumption_electrofuel / fuel_lhv * kerosene_market_price) / 1000000
        )
        self.df.loc[:, "electrofuel_cost_premium"] = electrofuel_cost_premium

        # Average cost per L
        electrofuel_avg_cost_per_l = (
            electrofuel_total_cost / energy_consumption_electrofuel * fuel_lhv * 1000000
        )
        self.df.loc[:, "electrofuel_avg_cost_per_l"] = electrofuel_avg_cost_per_l

        # Abatement cost in €/tCO2e (= overcost for a ton of biofuel/avoided emissions)
        carbon_abatement_cost_electrofuel = (
            electrofuel_cost_premium
            * 1000000
            / energy_consumption_electrofuel
            / (kerosene_emission_factor - electrofuel_emission_factor)
            * 1000000
        )
        self.df.loc[:, "carbon_abatement_cost_electrofuel"] = carbon_abatement_cost_electrofuel

        electrofuel_carbon_tax = (
            energy_consumption_electrofuel
            * electrofuel_emission_factor
            / 1000000
            * carbon_tax
            / 1000000
        )
        # M€
        self.df.loc[:, "electrofuel_carbon_tax"] = electrofuel_carbon_tax

        electrofuel_mfsp_carbon_tax_supplement = (
            carbon_tax * electrofuel_emission_factor / 1000000 * fuel_lhv
        )
        # €/L
        self.df.loc[
            :, "electrofuel_mfsp_carbon_tax_supplement"
        ] = electrofuel_mfsp_carbon_tax_supplement

        return (
            electrofuel_plant_building_scenario,
            electrofuel_plant_building_cost,
            electrofuel_total_cost,
            electrofuel_capex_cost,
            electrofuel_opex_cost,
            electrofuel_elec_cost,
            electrofuel_co2_cost,
            electrofuel_cost_premium,
            electrofuel_avg_cost_per_l,
            carbon_abatement_cost_electrofuel,
            electrofuel_carbon_tax,
            electrofuel_mfsp_carbon_tax_supplement,
        )

    @staticmethod
    def electrofuel_computation(
        electrofuel_eis_capex: pd.Series = pd.Series(dtype="float64"),
        electrofuel_eis_fixed_opex: pd.Series = pd.Series(dtype="float64"),
        electrofuel_eis_var_opex: pd.Series = pd.Series(dtype="float64"),
        # electrofuel_eis_specific_energy: pd.Series = pd.Series(dtype="float64"),
        electrolysis_efficiency: pd.Series = pd.Series(dtype="float64"),
        electrofuel_hydrogen_efficiency: pd.Series = pd.Series(dtype="float64"),
        electricity_market_price: pd.Series = pd.Series(dtype="float64"),
        energy_consumption_electrofuel: pd.Series = pd.Series(dtype="float64"),
        electricity_load_factor: float = 0.0,
        co2_market_price: pd.Series = pd.Series(dtype="float64"),
        electrofuel_eis_specific_co2: pd.Series = pd.Series(dtype="float64"),
    ):
        """
        Computes the yearly costs to respect a given hydrogen electrolysis production scenario.
        Capex in €/ton/day
        Fixed opex in  €/ton/day
        Variable opex in € per kg
        Specific electricity in kWh/kg --> Deprecated to harmonize with efficiency compytation, but might be reactivated later?
        Electricity market price in €/kWh
        Energy_consumption_hydrogen in MJ.
        Values returned are in M€ or t/day
        """
        # Convert hydrogen demand from MJ to tons.

        kerosene_specific_energy = 43
        demand_scenario = energy_consumption_electrofuel / kerosene_specific_energy / 1000
        plant_life = 30

        # Catching the indexes from the demand scenario
        indexes = demand_scenario.index

        # Additional plant capacity building needed (in ton/day output size) for a given year
        plant_building_scenario = pd.Series(np.zeros(len(indexes)), indexes)
        # Relative CAPEX cost to build the new facilities in M€2020
        plant_building_cost = pd.Series(np.zeros(len(indexes)), indexes)
        # Annual production in tons
        electrofuel_production = pd.Series(np.zeros(len(indexes)), indexes)

        # Annual cost and cost components of hydrogen production in M€
        electrofuel_total_cost = pd.Series(np.zeros(len(indexes)), indexes)
        electrofuel_capex_cost = pd.Series(np.zeros(len(indexes)), indexes)
        electrofuel_opex_cost = pd.Series(np.zeros(len(indexes)), indexes)
        electrofuel_elec_cost = pd.Series(np.zeros(len(indexes)), indexes)
        electrofuel_co2_cost = pd.Series(np.zeros(len(indexes)), indexes)

        # For each year o²f the demand scenario the demand is matched by the production
        for year in list(demand_scenario.index)[:-1]:
            # Production missing in year n+1 must be supplied by plant build in year n
            if electrofuel_production[year + 1] < demand_scenario[year + 1]:
                # Getting the production not matched by plants already commissioned by creating an electrolyzer with year data technical data

                electrofuel_cost = ElectrofuelCost.compute_electrofuel_year_lcop(
                    year,
                    electricity_market_price,
                    co2_market_price,
                    electricity_load_factor,
                    electrofuel_eis_capex,
                    electrofuel_eis_fixed_opex,
                    electrofuel_eis_var_opex,
                    electrolysis_efficiency,
                    electrofuel_hydrogen_efficiency,
                    electrofuel_eis_specific_co2,
                )

                # Getting the production not matched by plants already commissioned
                missing_production = demand_scenario[year + 1] - electrofuel_production[year + 1]

                # Converting the missing production to a capacity (in t/day)
                electrofuel_capacity_to_build = (
                    missing_production / 365.25 / electricity_load_factor
                )  # capacity to build in t/day production, taking into account load_factor

                electrolyser_capex_year = (
                    electrofuel_capacity_to_build * electrofuel_eis_capex[year] / 1000
                )  # electrolyzer capex is in €/kg/day or m€/ton/day ==> M€/ton/day
                plant_building_cost[year] = electrolyser_capex_year
                plant_building_scenario[year] = electrofuel_capacity_to_build  # in ton/day capacity

                # When production ends: either at the end of plant life or the end of the scenario;
                end_bound = min(list(demand_scenario.index)[-1], year + plant_life)

                # Adding new plant production to future years
                for i in range(year + 1, end_bound + 1):
                    electrofuel_total_cost[i] = (
                        electrofuel_total_cost[i]
                        + (missing_production * electrofuel_cost[i]["TOTAL"]) / 1000
                    )  # €/kg and production in tons => /1000 for M€
                    electrofuel_capex_cost[i] = (
                        electrofuel_capex_cost[i]
                        + (missing_production * electrofuel_cost[i]["CAPEX"]) / 1000
                    )  # M€
                    electrofuel_opex_cost[i] = (
                        electrofuel_opex_cost[i]
                        + (missing_production * electrofuel_cost[i]["FIX_OPEX"]) / 1000
                        + (missing_production * electrofuel_cost[i]["VAR_OPEX"]) / 1000
                    )  # M€
                    electrofuel_elec_cost[i] = (
                        electrofuel_elec_cost[i]
                        + (missing_production * electrofuel_cost[i]["ELECTRICITY"]) / 1000
                    )  # M€
                    electrofuel_co2_cost[i] = (
                        electrofuel_co2_cost[i]
                        + (missing_production * electrofuel_cost[i]["CO2"]) / 1000
                    )  # M€
                    electrofuel_production[i] = electrofuel_production[i] + missing_production

        # MOD -> Scaling down production for diminishing production scenarios.
        # Very weak model, assuming that production not anymore needed by aviation is used elsewhere in the industry.
        # Stranded asset literature could be valuable to model this better.
        # Proportional production scaling

        scaling_factor = demand_scenario / electrofuel_production

        if not all(scaling_factor.isna()):
            electrofuel_total_cost = electrofuel_total_cost * scaling_factor
            electrofuel_capex_cost = electrofuel_capex_cost * scaling_factor
            electrofuel_opex_cost = electrofuel_opex_cost * scaling_factor
            electrofuel_elec_cost = electrofuel_elec_cost * scaling_factor
            electrofuel_co2_cost = electrofuel_co2_cost * scaling_factor

        return (
            plant_building_scenario,
            plant_building_cost,
            electrofuel_total_cost,
            electrofuel_capex_cost,
            electrofuel_opex_cost,
            electrofuel_elec_cost,
            electrofuel_co2_cost,
        )

    @staticmethod
    def compute_electrofuel_year_lcop(
        base_year,
        electricity_market_price,
        co2_market_price,
        electricity_load_factor,
        electrofuel_capex,
        electrofuel_fixed_opex,
        electrofuel_var_opex,
        # electrofuel_specific_electricity,
        electrolysis_efficiency,
        electrofuel_hydrogen_efficiency,
        electrofuel_specific_co2,
    ):
        """
        This function computes the MFSP for electrofuel production for an plant commissioned at the base year.
        Costs are discounted to find a constant MFSP, excepted electricity, whose price is directly evolving
        from a year to another. Hydrogen production and fuel synthesis are considered a one-step process here.
        Capex in €/ (kg/day capacity)
        Fixed opex in €/ (kg/day capacity), on an annual basis
        Variable opex in € per kg
        Specific electricity in kWh/kg
        Specific CO2 in kg/kg
        Electricity market price in €/kWh
        Electrofuel price returned in €/kg
        """
        operational_time = 30
        hydrogen_prices = {}
        construction_time = 3

        discount = 0.10
        load_fact = min(0.95, electricity_load_factor)
        real_year_days = 365.25 * load_fact
        real_var_opex = electrofuel_var_opex[base_year] * real_year_days

        fuel_lhv = 35.3
        # https://www.engineeringtoolbox.com/fuels-higher-calorific-values-d_169.html
        # fuel density at 15 degrees
        fuel_density = 0.804

        dropin_specific_energy = fuel_lhv / fuel_density / 3.6  # kWh/kg

        electrofuel_specific_electricity = dropin_specific_energy / (
            electrolysis_efficiency * electrofuel_hydrogen_efficiency
        )

        cap_cost_npv = 0
        fix_op_cost_npv = 0
        var_op_cost_npv = 0
        hydrogen_npv = 0

        # Construction
        for i in range(0, construction_time):
            # The construction is supposed to span over x years, with a uniform cost repartition
            cap_cost_npv += (electrofuel_capex[base_year] / construction_time) / (1 + discount) ** (
                i
            )

        # commissioning
        for i in range(construction_time, operational_time + construction_time):
            fix_op_cost_npv += electrofuel_fixed_opex[base_year] / (1 + discount) ** (i)
            var_op_cost_npv += real_var_opex / (1 + discount) ** (i)
            hydrogen_npv += real_year_days / (1 + discount) ** (i)

        cap_cost_lc = cap_cost_npv / hydrogen_npv
        fix_op_cost_lc = fix_op_cost_npv / hydrogen_npv
        var_op_cost_lc = var_op_cost_npv / hydrogen_npv

        end_bound = min(max(electricity_market_price.index), operational_time + base_year)

        for year in range(base_year, end_bound + 1):
            elec_price = electricity_market_price[year]
            elec_cost = elec_price * electrofuel_specific_electricity[base_year]
            CO2_cost = electrofuel_specific_co2[base_year] * co2_market_price[year]
            hydrogen_prices[year] = {
                "TOTAL": cap_cost_lc + var_op_cost_lc + fix_op_cost_lc + elec_cost + CO2_cost,
                "CAPEX": cap_cost_lc,
                "FIX_OPEX": fix_op_cost_lc,
                "VAR_OPEX": var_op_cost_lc,
                "ELECTRICITY": elec_cost,
                "CO2": CO2_cost,
            }

        return hydrogen_prices


class ElectrofuelCapex(AeromapsModel):
    def __init__(self, name="electrofuel_capex", *args, **kwargs):
        super().__init__(name, *args, **kwargs)

    def compute(
        self,
        electrofuel_capex_2020: float = 0.0,
        electrofuel_capex_2030: float = 0.0,
        electrofuel_capex_2040: float = 0.0,
        electrofuel_capex_2050: float = 0.0,
    ) -> Tuple[pd.Series]:
        """electrofuel capital expenditures at eis using interpolation functions"""
        # FT MSW
        reference_values_capex = [
            electrofuel_capex_2020,
            electrofuel_capex_2030,
            electrofuel_capex_2040,
            electrofuel_capex_2050,
        ]

        reference_years = [2020, 2030, 2040, 2050]

        capex_function = interp1d(reference_years, reference_values_capex, kind="linear")
        for k in range(self.prospection_start_year, self.end_year + 1):
            self.df.loc[k, "electrofuel_eis_capex"] = capex_function(k)

        electrofuel_eis_capex = self.df.loc[:, "electrofuel_eis_capex"]

        return electrofuel_eis_capex


class ElectrofuelFixedOpex(AeromapsModel):
    def __init__(self, name="electrofuel_fixed_opex", *args, **kwargs):
        super().__init__(name, *args, **kwargs)

    def compute(
        self,
        electrofuel_fixed_opex_2020: float = 0.0,
        electrofuel_fixed_opex_2030: float = 0.0,
        electrofuel_fixed_opex_2040: float = 0.0,
        electrofuel_fixed_opex_2050: float = 0.0,
    ) -> Tuple[pd.Series]:
        """electrofuel fixed operational expenditures at entry into service using interpolation functions"""
        # FT MSW
        reference_values_fixed_opex = [
            electrofuel_fixed_opex_2020,
            electrofuel_fixed_opex_2030,
            electrofuel_fixed_opex_2040,
            electrofuel_fixed_opex_2050,
        ]

        reference_years = [2020, 2030, 2040, 2050]

        fixed_opex_function = interp1d(reference_years, reference_values_fixed_opex, kind="linear")
        for k in range(self.prospection_start_year, self.end_year + 1):
            self.df.loc[k, "electrofuel_eis_fixed_opex"] = fixed_opex_function(k)

        electrofuel_eis_fixed_opex = self.df.loc[:, "electrofuel_eis_fixed_opex"]

        return electrofuel_eis_fixed_opex


class ElectrofuelVarOpex(AeromapsModel):
    def __init__(self, name="electrofuel_var_opex", *args, **kwargs):
        super().__init__(name, *args, **kwargs)

    def compute(
        self,
        electrofuel_var_opex_2020: float = 0.0,
        electrofuel_var_opex_2030: float = 0.0,
        electrofuel_var_opex_2040: float = 0.0,
        electrofuel_var_opex_2050: float = 0.0,
    ) -> Tuple[pd.Series]:
        """electrofuel variable operational expenditures at entry into service using interpolation functions"""
        # FT MSW
        reference_values_var_opex = [
            electrofuel_var_opex_2020,
            electrofuel_var_opex_2030,
            electrofuel_var_opex_2040,
            electrofuel_var_opex_2050,
        ]

        reference_years = [2020, 2030, 2040, 2050]

        var_opex_function = interp1d(reference_years, reference_values_var_opex, kind="linear")
        for k in range(self.prospection_start_year, self.end_year + 1):
            self.df.loc[k, "electrofuel_eis_var_opex"] = var_opex_function(k)

        electrofuel_eis_var_opex = self.df.loc[:, "electrofuel_eis_var_opex"]

        return electrofuel_eis_var_opex


########## Deprecated for the time being, might be reactivated. ####################
#
# class ElectrofuelSpecificElectricity(AeromapsModel):
#
#     # changement d'usage par rapport à CAST==> on utilise pas l'efficacité moyenne pour la cons d'élec, mais
#     # l'efficacité de chaque année de mise en service de l'eclectolyseur. Permet de faire des choses plus détaillées
#     #
#     def __init__(self, name="electrofuel_specific_electricity", *args, **kwargs):
#         super().__init__(name, *args, **kwargs)
#
#     def compute(
#             self,
#             electrofuel_specific_electricity_2020: float = 0.0,
#             electrofuel_specific_electricity_2030: float = 0.0,
#             electrofuel_specific_electricity_2040: float = 0.0,
#             electrofuel_specific_electricity_2050: float = 0.0,
#     ) -> Tuple[pd.Series]:
#         """electrofuel efficiency at eis using interpolation functions"""
#         # FT MSW
#         reference_values_specific_electricity = [
#             electrofuel_specific_electricity_2020,
#             electrofuel_specific_electricity_2030,
#             electrofuel_specific_electricity_2040,
#             electrofuel_specific_electricity_2050
#         ]
#
#         reference_years = [2020, 2030, 2040, 2050]
#
#         specific_electricity_function = interp1d(
#             reference_years, reference_values_specific_electricity, kind="linear"
#         )
#         for k in range(self.prospection_start_year, self.end_year + 1):
#             self.df.loc[
#                 k, "electrofuel_eis_specific_electricity"
#             ] = specific_electricity_function(k)
#
#         electrofuel_eis_specific_electricity = self.df.loc[:, "electrofuel_eis_specific_electricity"]
#
#         return (
#             electrofuel_eis_specific_electricity
#         )
####################################################################################


class ElectrofuelSpecificCo2(AeromapsModel):

    # changement d'usage par rapport à CAST==> on utilise pas l'efficacité moyenne pour la cons d'élec, mais
    # l'efficacité de chaque année de mise en service de l'eclectolyseur. Permet de faire des choses plus détaillées.
    def __init__(self, name="electrofuel_specific_co2", *args, **kwargs):
        super().__init__(name, *args, **kwargs)

    def compute(
        self,
        electrofuel_specific_co2_2020: float = 0.0,
        electrofuel_specific_co2_2030: float = 0.0,
        electrofuel_specific_co2_2040: float = 0.0,
        electrofuel_specific_co2_2050: float = 0.0,
    ) -> Tuple[pd.Series]:
        """electrofuel efficiency at eis using interpolation functions"""
        # FT MSW
        reference_values_specific_co2 = [
            electrofuel_specific_co2_2020,
            electrofuel_specific_co2_2030,
            electrofuel_specific_co2_2040,
            electrofuel_specific_co2_2050,
        ]

        reference_years = [2020, 2030, 2040, 2050]

        specific_co2_function = interp1d(
            reference_years, reference_values_specific_co2, kind="linear"
        )
        for k in range(self.prospection_start_year, self.end_year + 1):
            self.df.loc[k, "electrofuel_eis_specific_co2"] = specific_co2_function(k)

        electrofuel_eis_specific_co2 = self.df.loc[:, "electrofuel_eis_specific_co2"]

        return electrofuel_eis_specific_co2



# @Time : 13/03/2023 10:46
# @Author : a.salgas
# @File : scenario_cost.py
# @Software: PyCharm

from typing import Tuple

import pandas as pd
from aeromaps.models.base import AeromapsModel


class NonDiscountedScenarioCost(AeromapsModel):
    def __init__(self, name="non_discounted_scenario_cost", *args, **kwargs):
        super().__init__(name, *args, **kwargs)

    def compute(
        self,
        kerosene_cost: pd.Series = pd.Series(dtype="float64"),
        biofuel_cost_hefa_fog: pd.Series = pd.Series(dtype="float64"),
        biofuel_cost_hefa_others: pd.Series = pd.Series(dtype="float64"),
        biofuel_cost_ft_others: pd.Series = pd.Series(dtype="float64"),
        biofuel_cost_ft_msw: pd.Series = pd.Series(dtype="float64"),
        biofuel_cost_atj: pd.Series = pd.Series(dtype="float64"),
        electrofuel_total_cost: pd.Series = pd.Series(dtype="float64"),
        total_hydrogen_supply_cost: pd.Series = pd.Series(dtype="float64"),
        biofuel_cost_premium_hefa_fog: pd.Series = pd.Series(dtype="float64"),
        biofuel_cost_premium_hefa_others: pd.Series = pd.Series(dtype="float64"),
        biofuel_cost_premium_ft_others: pd.Series = pd.Series(dtype="float64"),
        biofuel_cost_premium_ft_msw: pd.Series = pd.Series(dtype="float64"),
        biofuel_cost_premium_atj: pd.Series = pd.Series(dtype="float64"),
        h2_cost_premium_electrolysis: pd.Series = pd.Series(dtype="float64"),
        h2_cost_premium_gas_ccs: pd.Series = pd.Series(dtype="float64"),
        h2_cost_premium_gas: pd.Series = pd.Series(dtype="float64"),
        h2_cost_premium_coal_ccs: pd.Series = pd.Series(dtype="float64"),
        h2_cost_premium_coal: pd.Series = pd.Series(dtype="float64"),
        electrofuel_cost_premium: pd.Series = pd.Series(dtype="float64"),
    ) -> Tuple[pd.Series, pd.Series, pd.Series]:
        # Compute the total energy expenses of the scenario
        non_discounted_energy_expenses = (
            kerosene_cost.fillna(0)
            + biofuel_cost_hefa_fog.fillna(0)
            + biofuel_cost_hefa_others.fillna(0)
            + biofuel_cost_ft_others.fillna(0)
            + biofuel_cost_ft_msw.fillna(0)
            + biofuel_cost_atj.fillna(0)
            + electrofuel_total_cost.fillna(0)
            + total_hydrogen_supply_cost.fillna(0)
        )

        # Compute the total cost premium compared to the "business as usual"
        non_discounted_energy_cost_premium = (
            biofuel_cost_premium_atj.fillna(0)
            + biofuel_cost_premium_ft_msw.fillna(0)
            + biofuel_cost_premium_ft_others.fillna(0)
            + biofuel_cost_premium_hefa_others.fillna(0)
            + biofuel_cost_premium_hefa_fog.fillna(0)
            + h2_cost_premium_electrolysis.fillna(0)
            + h2_cost_premium_gas_ccs.fillna(0)
            + h2_cost_premium_gas.fillna(0)
            + h2_cost_premium_coal_ccs.fillna(0)
            + h2_cost_premium_coal.fillna(0)
            + electrofuel_cost_premium.fillna(0)
        )

        # Compute the business as usual energy expenses
        non_discounted_BAU_energy_expenses = (
            non_discounted_energy_expenses - non_discounted_energy_cost_premium
        )

        self.df.loc[:, "non_discounted_energy_expenses"] = non_discounted_energy_expenses
        self.df.loc[:, "non_discounted_energy_cost_premium"] = non_discounted_energy_cost_premium
        self.df.loc[:, "non_discounted_BAU_energy_expenses"] = non_discounted_BAU_energy_expenses

        return (
            non_discounted_energy_expenses,
            non_discounted_energy_cost_premium,
            non_discounted_BAU_energy_expenses,
        )


class DicountedScenarioCost(AeromapsModel):
    def __init__(self, name="discounted_scenario_cost", *args, **kwargs):
        super().__init__(name, *args, **kwargs)

    def compute(
        self,
        social_discount_rate: float = 0.0,
        kerosene_cost: pd.Series = pd.Series(dtype="float64"),
        biofuel_cost_hefa_fog: pd.Series = pd.Series(dtype="float64"),
        biofuel_cost_hefa_others: pd.Series = pd.Series(dtype="float64"),
        biofuel_cost_ft_others: pd.Series = pd.Series(dtype="float64"),
        biofuel_cost_ft_msw: pd.Series = pd.Series(dtype="float64"),
        biofuel_cost_atj: pd.Series = pd.Series(dtype="float64"),
        electrofuel_total_cost: pd.Series = pd.Series(dtype="float64"),
        total_hydrogen_supply_cost: pd.Series = pd.Series(dtype="float64"),
    ) -> Tuple[pd.Series]:

        for k in range(self.prospection_start_year, self.end_year + 1):

            # Compute the discounter at year k
            discount_k = (1 + social_discount_rate) ** (k - self.prospection_start_year)

            # Compute the discounted expenses for each energy pathway
            kerosene_discounted = kerosene_cost.fillna(0)[k] / discount_k
            biofuel_hefa_fog_discounted = biofuel_cost_hefa_fog.fillna(0)[k] / discount_k
            biofuel_hefa_others_discounted = biofuel_cost_hefa_others.fillna(0)[k] / discount_k
            biofuel_ft_others_discounted = biofuel_cost_ft_others.fillna(0)[k] / discount_k
            biofuel_ft_msw_discounted = biofuel_cost_ft_msw.fillna(0)[k] / discount_k
            biofuel_atj_discounted = biofuel_cost_atj.fillna(0)[k] / discount_k
            electrofuel_discounted = electrofuel_total_cost.fillna(0)[k] / discount_k
            hydrogen_discounted = total_hydrogen_supply_cost.fillna(0)[k] / discount_k

            self.df.loc[k, "discounted_energy_expenses"] = (
                kerosene_discounted
                + biofuel_atj_discounted
                + biofuel_ft_msw_discounted
                + biofuel_ft_others_discounted
                + biofuel_hefa_others_discounted
                + biofuel_hefa_fog_discounted
                + electrofuel_discounted
                + hydrogen_discounted
            )

        discounted_energy_expenses = self.df.loc[:, "discounted_energy_expenses"]

        return discounted_energy_expenses



from dataclasses import dataclass, asdict, field
from json import dump, load
import pandas as pd

from dacite import from_dict

from aeromaps.utils.functions import _dict_from_json

from aeromaps.models.sustainability_assessment.parameters import SustainabilityAssessmentParameters
from aeromaps.models.impacts.parameters import ImpactsParameters
from aeromaps.models.air_transport.parameters import AirTransportParameters


@dataclass
class YearParameters:
    historic_start_year: int = 2000
    prospection_start_year: int = 2020
    end_year: int = 2050


@dataclass
class AllParameters(
    YearParameters, SustainabilityAssessmentParameters, ImpactsParameters, AirTransportParameters
):
    @property
    def __dict__(self):
        parameters_dict = asdict(self)

        for key, value in parameters_dict.items():
            if isinstance(value, pd.Series):
                parameters_dict[key] = list(value)

        return parameters_dict

    def write_json(self, file_name="parameters.json"):
        with open(file_name, "w", encoding="utf-8") as f:
            dump(self.__dict__, f, ignore_nan=True, ensure_ascii=False, indent=4)

    @staticmethod
    def read_json(file_name="parameters.json"):

        parameters_dict = _dict_from_json(file_name=file_name)

        return from_dict(data_class=AllParameters, data=parameters_dict)


all_parameters = AllParameters()
